def yield_logs(group, stream, start_time=None, session=None, client=None):
    """
    yields log lines from the specified group and stream from the specified
    start time onwards
    """
    client = client or session.client('logs')
    kwargs = {
        'logGroupName': group,
        'logStreamName': stream,
        'startFromHead': True,
    }
    if start_time:
        kwargs['startTime'] = start_time + 1
    next_token = 'sesame'
    while next_token:
        last_token = next_token
        response = client.get_log_events(**kwargs)
        if not response['events']:
            break
        yield from response['events']
        next_token = response.get('nextForwardToken')
        if next_token == last_token:
            break
        kwargs['nextToken'] = next_token



import json
import os
from datetime import datetime
from botocore.exceptions import WaiterError
from botocore.exceptions import ClientError
from raft.tasks import task

from ...base.utils import notice_end, notice, print_table, confirm
from ...base.utils import load_yaml
from ..base import AwsTask
from ..base import post_to_slack


__all__ = [
    'upsert',
    'delete_stack',
    'test',
    'sync_template',
    'show_parameters',
    'push_template_to_s3',
    'tail_stack',
]


def cloud_formation_bucket(session):
    s3 = session.resource('s3')
    for x in s3.buckets.all():
        if x.name.startswith('cf-templates') and x.name.endswith(session.region_name):
            return x
    return None


def list_events(stack_name, stack, last_event):
    try:
        events = list(stack.events.all())[::-1]
        n = 0
        found = False
        for x in events:
            if x.id == last_event:
                found = True
                break
            n += 1
        if not found:
            n = -1
        for x in events[n + 1:]:
            status = x.resource_status
            if status.startswith('UPDATE_COMPLETE_'):
                status = status[16:]
            if x.resource_status_reason:
                print(f'{x.logical_resource_id:30} {status:20}'
                      f' {x.resource_status_reason}')
            else:
                print(f'{x.logical_resource_id:30} {status:20}')
        if events:
            event = events[-1]
            resource_status = event.resource_status
            logical_resource_id = event.logical_resource_id
            if logical_resource_id == stack_name and resource_status.endswith('_COMPLETE'):
                return False
            last_event = event.id
    except:
        pass
    return last_event


def wait_for_event(session, stack_name, event_name, last_event=None, n_max=720, delay=3):
    cloudformation = session.resource('cloudformation')
    stack = cloudformation.Stack(stack_name)
    client = session.client('cloudformation')
    waiter = client.get_waiter(event_name)
    waiter.config.delay = delay
    waiter.config.max_attempts = 1
    n = 0
    while True:
        n += 1
        if n > n_max:
            break
        try:
            waiter.wait(StackName=stack_name)
            list_events(stack_name, stack, last_event)
            break
        except Exception:
            last_event = list_events(stack_name, stack, last_event)
            if last_event is False:
                return


def load_cfn_parameters(ctx, initial=None, prefix=None):
    params = {}
    global_params = os.path.join(prefix, 'params/global.yml')
    files = [ global_params ]
    try:
        files = ctx.aws.global_parameter_files + files
    except AttributeError:
        pass
    for x in files:
        notice(f'loading {os.path.basename(x)}')
        loaded = False
        if os.path.exists(x):
            try:
                y = load_yaml(x)
                if y:
                    params.update(y)
                notice_end()
                loaded = True
            except UnicodeDecodeError:
                pass
        if not loaded:
            notice_end('not found, skipping')
    params.update(initial)
    return params


def push_template_to_s3(stack_name, filename, session):
    """
    filename can be either a filename (/path/to/file or a buffer like io.BytesIO)
    """
    notice('getting cfn bucket')
    bucket = cloud_formation_bucket(session)
    if bucket is None:
        notice_end('not found')
        if isinstance(filename, str):
            with open(filename, 'r') as f:
                data = f.read()
        else:
            data = filename.getvalue()
        return None, data

    notice_end(bucket.name)
    key = datetime.now().strftime(f'{stack_name}/%Y.%m.%d@%H.%I.%S.yml')
    notice('pushing template to s3')
    if isinstance(filename, str):
        with open(filename, 'rb') as data:
            bucket.upload_fileobj(data, key)
    else:
        bucket.upload_fileobj(filename, key)

    template_url = (
        f'https://s3.{session.region_name}.amazonaws.com/'
        f'{bucket.name}/{key}' )
    notice_end(template_url)
    return template_url, None


def template(url, body):
    if url:
        return { 'TemplateURL': url }
    return { 'TemplateBody': body }


def fill_parameter_values(
        ctx, stack_name, template_url, template_body, stack_parameters,
        b_create, session, prefix):
    client = session.client('cloudformation')

    result = client.validate_template(**template(template_url, template_body))
    stack_parameters = stack_parameters or {}
    current_parameters = set()
    if not b_create:
        try:
            stacks = client.describe_stacks(StackName=stack_name)
            for x in stacks['Stacks'][0]['Parameters']:
                current_parameters.add(x['ParameterKey'])
        except:
            pass
    result = result['Parameters']
    local_values = load_cfn_parameters(ctx, stack_parameters, prefix=prefix)
    values = []
    for parameter in result:
        x = parameter['ParameterKey']
        if x in local_values:
            values.append({
                'ParameterKey': x,
                'ParameterValue': local_values[x],
            })
        elif x in current_parameters:
            values.append({
                'ParameterKey': x,
                'UsePreviousValue': True,
            })
    return values


def parameters(ctx, session, prefix, stack_name, filename, b_create=False):
    """
    :param session: a boto3 session
    :param prefix: the dir name
    :param stack_name: the stack name
    :param filename: the /path/to the yaml cloudformation file
     :type filename: str
    :param b_create: whether the parameters should be specified for a
           a create operation or for an update operation
     :type b_create: bool
    :return: a tuple with the yaml file contents as well as the parameter array
     :rtype: (str, str, list[dict])
    """
    template_url, template_body = push_template_to_s3(stack_name, filename, session)
    values = fill_parameter_values(
        ctx, stack_name, template_url, template_body, None,
        b_create, session, prefix)
    return template_url, template_body, values


def parameterized(ctx, session, prefix, stack_name, filename, b_create=False):
    """
    :param session: a boto3 session that can be used to create clients
    :param prefix: the directory name that precedes the `templates` directory
    :param stack_name: the stack name
    :param filename: the /path/to the yaml cloudformation file
     :type filename: str
    :param b_create: whether the parameters should be specified for a
           a create operation or for an update operation
     :type b_create: bool
    :return: a tuple with the yaml file contents as well as the parameter array
     :rtype: (str, bytes, list[dict])
    """
    stack_definition = load_yaml(filename)
    cfn_filename = stack_definition['template']
    template_dir = os.path.join(prefix, 'cfn')
    cfn_filename = os.path.join(template_dir, f'{cfn_filename}.yml')
    stack_parameters = stack_definition['parameters']
    template_url, template_body = push_template_to_s3(
        stack_name, cfn_filename, session)
    values = fill_parameter_values(
        ctx, stack_name, template_url, template_body, stack_parameters,
        b_create, session, prefix)

    return template_url, template_body, values


@task(klass=AwsTask)
def upsert(
        ctx, stack_name, filename=None, force_delete=False,
        timeout=60, disable_rollback=False,
        session=None, prefix='', **kwargs):
    """
    Creates/Updates the cloud formation named <stack_name>
    with the <filename> template
    :param stack_name: The name of the cloud formation stack
    :param filename: The /path/to/the/cloudformation.template
    :param force_delete: <true|false> If the stack exists whether to delete the stack and recreate it
    :param int timeout: Number of minutes for the timeout for cloudformation
    :return:
    """
    cfn = session.resource('cloudformation')
    stack = cfn.Stack(stack_name)
    stack_exists = False
    prefix = prefix or ctx.upholstery.prefix
    original_filename = filename
    params_dir = os.path.join(prefix, 'params')
    template_dir = os.path.join(prefix, 'cfn')
    if not filename:
        filename = os.path.join(params_dir, f'{stack_name}.yml')
    if filename.startswith(params_dir) and os.path.exists(filename):
        parameterized_upsert(
            ctx,
            session,
            prefix,
            stack_name,
            original_filename,
            force_delete,
            timeout)
        return
    if not original_filename:
        filename = os.path.join(template_dir, f'{stack_name}.yml')
    if isinstance(force_delete, str):
        force_delete = force_delete.lower() in [ 'yes', 'true' ]
    try:
        stack.load()
        stack_exists = True
    except:
        pass
    client = session.client('cloudformation')
    if stack_exists and force_delete:
        delete_stack(ctx, stack_name, session=session, prefix=prefix)
        stack_exists = False
    last_event = None
    event_name = 'stack_create_complete'
    post_to_slack(
        ctx, session, 'infrastructure',
        f'[*{prefix}/{stack_name}*] upserting')
    template_url, template_body, values = parameters(
        ctx, session, prefix, stack_name, filename, not stack_exists)
    if not stack_exists:
        notice(f'creating stack [{stack_name}]')
        client.create_stack(
            StackName=stack_name,
            Capabilities=[ 'CAPABILITY_NAMED_IAM' ],
            Parameters=values,
            # OnFailure='DELETE',
            DisableRollback=disable_rollback,
            TimeoutInMinutes=timeout,
            **template(template_url, template_body)
        )
    else:
        notice(f'updating stack [{stack_name}]')
        last_event = next(iter(stack.events.all())).id
        try:
            stack.update(
                Capabilities=[ 'CAPABILITY_NAMED_IAM' ],
                DisableRollback=disable_rollback,
                Parameters=values,
                **template(template_url, template_body)
            )
        except ClientError as ex:
            response = ex.response
            error = response['Error']
            m = error['Message']
            if 'No updates are to be performed' in f'{ex}':
                notice_end('nothing to be done')
                return
            rollback_complete = error['Code'] == 'ValidationError'
            rollback_complete = rollback_complete and 'ROLLBACK_COMPLETE' in m
            if rollback_complete:
                notice_end('state is rollback_complete, nothing doing')
                return
            raise ClientError from ex
        event_name = 'stack_update_complete'

    notice_end('initiated')
    wait_for_event(session, stack_name, event_name, last_event)
    notice(f'{stack_name} upsert')
    notice_end('complete')
    post_to_slack(
        ctx, session, 'infrastructure',
        f'[*{prefix}/{stack_name}*] upserted')
    if stack_name.endswith('-ami'):
        from ..ami_helpers import wait_for_ami
        wait_for_ami(ctx, stack_name[:-4], session=session, prefix=prefix)


def parameterized_upsert(
        ctx, session, prefix, stack_name, filename=None,
        force_delete=False, timeout=30):
    cloudformation = session.resource('cloudformation')
    stack = cloudformation.Stack(stack_name)
    stack_exists = False
    if not filename:
        filename = os.path.join(prefix, 'params', stack_name + '.yml')
    if not os.path.exists(filename):
        notice_end(f'Could not find file named {filename}')
    if force_delete and isinstance(force_delete, str):
        force_delete = force_delete.lower() in [ 'yes', 'true' ]
    try:
        stack.load()
        stack_exists = True
    except:
        pass
    client = session.client('cloudformation')
    if stack_exists and force_delete:
        delete_stack(ctx, stack_name, session=session, prefix=prefix)
        stack_exists = False
    last_event = None
    event_name = 'stack_create_complete'
    post_to_slack(
        ctx, session, 'infrastructure',
        f'[*{prefix}/{stack_name}*] upserting',)
    template_url, template_body, values = parameterized(
        ctx, session, prefix, stack_name, filename, not stack_exists)
    if not stack_exists:
        notice(f'Creating stack {stack_name}')
        client.create_stack(
            StackName=stack_name,
            Capabilities=[ 'CAPABILITY_NAMED_IAM' ],
            Parameters=values,
            # OnFailure='DELETE',
            TimeoutInMinutes=timeout,
            **template(template_url, template_body)
        )
    else:
        notice(f'Updating stack {stack_name}')
        last_event = next(iter(stack.events.all())).id
        try:
            stack.update(
                Capabilities=[ 'CAPABILITY_NAMED_IAM' ],
                Parameters=values,
                **template(template_url, template_body),
            )
            event_name = 'stack_update_complete'
        except ClientError as ex:
            message = f'{ex}'
            if 'No updates' in message:
                notice_end('no updates')
                post_to_slack(
                    ctx, session, 'infrastructure',
                    f'[*{prefix}/{stack_name}*] nothing done',)
                return
            post_to_slack(
                ctx, session, 'infrastructure',
                f'[*{prefix}/{stack_name}*] nothing done / error')
            raise ex

    notice_end('initiated')
    wait_for_event(session, stack_name, event_name, last_event)
    notice(f'{stack_name} upsert')
    notice_end('complete')
    post_to_slack(
        ctx, session, 'infrastructure',
        f'[*{prefix}/{stack_name}*] upserted')


def print_summary(changes):
    resource_changes = [ x['ResourceChange'] for x in changes if x['Type'] == 'Resource' ]
    header = [ 'resource', 'action', 'replacement', 'type', 'changes', ]
    rows = []
    for x in resource_changes:
        name = x.get('LogicalResourceId')
        name = name or x.get('PhysicalResourceId')
        type_ = x['ResourceType'].split('::')[-1]
        action = x['Action']
        replacement = x.get('Replacement', ' ')
        if action.lower() == 'modify':
            details = x.get('Details') or []
            rg = []
            for change in details:
                evaluation = change.get('Evaluation')
                if evaluation.lower() == 'dynamic':
                    rg.append(change.get('CausingEntity') or '')
                else:
                    t = change['Target'].get('Name')
                    if not t:
                        t = change['Target']['Attribute']
                    rg.append(t)
            rg = filter(None, rg)
            change_summary = ', '.join(rg)
            change_summary = change_summary or ' '
        else:
            change_summary = ' '

        fields = [ name, action, replacement, type_, change_summary ]
        rows.append(fields)
    print_table(header, rows)


@task(klass=AwsTask)
def test(
        ctx, stack_name, filename=None, verbose=None, session=None,
        prefix=None, **kwargs):
    """
    Creates/Updates the cloud formation named <stack_name>
    with the <filename> template
    :param session: boto3 session (passed in from decorator)
    :param prefix: passed in from partial
    :param stack_name: The name of the cloud formation stack
    :param filename: The /path/to/the/cloudformation.template
    :param verbose: specify the verbose flag to indicate that you want verbose output instead of
      summary output
    :return:
    """
    cloudformation = session.resource('cloudformation')
    stack = cloudformation.Stack(stack_name)
    stack_exists = False
    prefix = prefix or ctx.upholstery.prefix
    original_filename = filename
    params_dir = os.path.join(prefix, 'params')
    template_dir = os.path.join(prefix, 'cfn')
    if not filename:
        filename = os.path.join(params_dir, f'{stack_name}.yml')
    if filename.startswith(params_dir) and os.path.exists(filename):
        parameterized_test(
            ctx, session, prefix, stack_name, original_filename, verbose)
        return
    if not original_filename:
        filename = os.path.join(template_dir, f'{stack_name}.yml')
    try:
        stack.load()
        stack_exists = True
    except:
        pass
    client = session.client('cloudformation')
    name = datetime.now().strftime(stack_name + '-%Y%m%d-%H%M')
    change_set_type = 'UPDATE' if stack_exists else 'CREATE'
    template_url, template_body, values = parameters(
        ctx, session, prefix, stack_name, filename, not stack_exists)
    notice(f'testing changes to [{stack_name}]')
    client.create_change_set(
        ChangeSetName=name,
        ChangeSetType=change_set_type,
        StackName=stack_name,
        Capabilities=[ 'CAPABILITY_NAMED_IAM' ],
        Parameters=values,
        **template(template_url, template_body),
    )
    waiter = client.get_waiter('change_set_create_complete')
    try:
        waiter.wait(ChangeSetName=name, StackName=stack_name)
        notice_end()
    except WaiterError:
        notice_end('✘ (timed out)')
        return
    data = client.describe_change_set(
        ChangeSetName=name,
        StackName=stack_name,
    )
    if data:
        reason = data.get('StatusReason', '')
        if reason.startswith(
            "The submitted information didn't contain changes"
        ):
            notice_end('nothing to be done')
            return
        changes = data['Changes']
        if verbose:
            print(json.dumps(changes, indent=2))
        else:
            print_summary(changes)

        client.delete_change_set(
            ChangeSetName=name,
            StackName=stack_name,)
    notice_end('done')


def parameterized_test(ctx, session, prefix, stack_name,
                       filename=None, verbose=None):
    """
    <stack_name>,<filename>,verbose -- Creates/Updates the cloud formation named <stack_name>
    with the <filename> template
    :param session: the boto3 session passed in from the decorator
    :param prefix: the prefix (e.g., dbuy or leaseco) passed in from partial
    :param stack_name: The name of the cloud formation stack
    :param filename: The /path/to/the/cloudformation.template
    :param verbose: specify the verbose flag to indicate that you want verbose output instead of
      summary output
    :return:
    """
    cloudformation = session.resource('cloudformation')
    stack = cloudformation.Stack(stack_name)
    stack_exists = False
    params_dir = os.path.join(prefix, 'params')
    if not filename:
        filename = os.path.join(params_dir, stack_name + '.yml')
    if not os.path.exists(filename):
        notice(f'Could not find [{filename}]')
        notice_end('done')
        return
    try:
        stack.load()
        stack_exists = True
    except:
        pass
    client = session.client('cloudformation')
    name = datetime.now().strftime(stack_name + '-%Y%m%d-%H%M')
    change_set_type = 'UPDATE' if stack_exists else 'CREATE'
    template_url, template_body, values = parameterized(
        ctx, session, prefix, stack_name, filename, not stack_exists)
    notice(f'testing changes to [{stack_name}]')
    client.create_change_set(
        ChangeSetName=name,
        ChangeSetType=change_set_type,
        StackName=stack_name,
        Capabilities=[ 'CAPABILITY_NAMED_IAM' ],
        Parameters=values,
        **template(template_url, template_body),
    )
    try:
        waiter = client.get_waiter('change_set_create_complete')
        waiter.wait(ChangeSetName=name, StackName=stack_name)
        data = client.describe_change_set(
            ChangeSetName=name,
            StackName=stack_name,
        )
        changes = data['Changes']
        if verbose:
            print(f'{json.dumps(changes, indent=2)}')
        else:
            print_summary(changes)
        notice_end('done')
    except WaiterError:
        notice_end('up to date')
    client.delete_change_set(
        ChangeSetName=name,
        StackName=stack_name,
    )


@task(klass=AwsTask)
def delete_stack(ctx, stack_name, session=None, prefix='', **kwargs):
    """
    <stack_name> -- Deletes the specified cloud formation stack named <stack_name>
    :param ctx: the convocation context
    :param stack_name: The name of the cloud formation stack
    :param session: the boto3 session passed in from the decorator
    :param prefix: the prefix passed in from partial
    """
    cloudformation = session.resource('cloudformation')
    stack = cloudformation.Stack(stack_name)
    stack_exists = False
    if not prefix:
        prefix = ctx.upholstery.prefix
    try:
        stack.load()
        stack_exists = True
    except:
        pass
    if stack_exists:
        if not confirm(f'You are requesting deletion of stack [{stack_name}]'):
            return
        post_to_slack(
            ctx, session, 'infrastructure',
            f'[*{prefix}/{stack_name}*] deleting')
        last_event = next(iter(stack.events.all())).id
        notice(f'Deleting existing stack {stack_name}')
        stack.delete()
        notice_end()
        wait_for_event(session, stack_name, 'stack_delete_complete', last_event)
        notice(f'{stack_name} deletion')
        notice_end('complete')
        post_to_slack(
            ctx, session, 'infrastructure',
            f'[*{prefix}/{stack_name}*] deleted')


@task(klass=AwsTask)
def sync_template(ctx, stack_name, with_parameters=False,
                  session=None, prefix=None, **kwargs):
    """
    pulls the latest template from cfn and saves it to the cfn directory
    """
    from convocations.base.utils import yaml_serializer
    cfn = session.client('cloudformation')
    response = cfn.get_template(StackName=stack_name)
    body = response['TemplateBody']
    prefix = prefix or ctx.upholstery.prefix
    filename = os.path.join(prefix, 'cfn', f'{stack_name}.yml')
    with open(filename, 'w') as f:
        if isinstance(body, dict):
            body = json.loads(json.dumps(body))
            yaml_serializer().dump(body, f)
        else:
            f.write(body)
            f.write('\n')
    if with_parameters:
        parameter_response = cfn.describe_stacks(StackName=stack_name)
        p = parameter_response['Stacks'][0]['Parameters']
        filename = os.path.join(prefix, 'params', f'{stack_name}.yml')
        data = dict(template=stack_name, parameters={
            x['ParameterKey']: x['ParameterValue']
            for x in p
        })
        with open(filename, 'w') as f:
            data = json.loads(json.dumps(data))
            yaml_serializer().dump(data, f)


@task(klass=AwsTask)
def show_parameters(ctx, name, session=None, **kwargs):
    cfn = session.client('cloudformation')
    result = cfn.describe_stacks(StackName=name)
    stack = result['Stacks'][0]
    mp = stack['Parameters']
    for x in mp:
        notice(x['ParameterKey'])
        notice_end(x['ParameterValue'])
    return mp


@task(klass=AwsTask)
def tail_stack(ctx, stack_name, event_name='stack_update_complete',
               n_max=360, delay=10,
               session=None, **kwargs):
    """
    tails all stack events for the given stack until event_name is reached
    delay is the delay between polling
    n_max is the max number of polls for events
    """
    last_event = None
    cloudformation = session.resource('cloudformation')
    stack = cloudformation.Stack(stack_name)
    stack.load()
    for x in list(stack.events.all())[:5]:
        last_event = x.id
    wait_for_event(session, stack_name, event_name, last_event, n_max, delay)



import json
import os
import webbrowser
from datetime import datetime
from getpass import getuser
from hashlib import sha1
from time import sleep
from time import time
from typing import Optional, Tuple, Any, List, Dict
from typing import Mapping

import boto3.session
import requests
from dateutil.parser import parse as parse_time
import pytz
from botocore.utils import SSOTokenLoader
from botocore.credentials import JSONFileCache
from botocore.exceptions import ClientError
from raft import Context
from raft.tasks import Task
from ..base.utils import notice, notice_end, get_context_value

g_cache_dir_noticed = False


def cache_dir():
    """
    gets the aws cache dir, which is user specific as ~/.aws/sso/cache
    and ensures that the directories exist with permissions 0700
    all the way down
    """
    global g_cache_dir_noticed
    if not g_cache_dir_noticed:
        notice('checking for cache dir')
    d = os.path.expanduser('~')
    dirs = [ '.aws', 'sso', 'cache' ]
    rg = []
    for x in dirs:
        d = os.path.join(d, x)
        rg.append(d)
    if not os.path.exists(d):
        for x in rg:
            os.makedirs(x, 0o700, exist_ok=True)
    if not g_cache_dir_noticed:
        notice_end(d, 'green')
        g_cache_dir_noticed = True
    return d


def client_credentials(sso_oidc, force=False):
    """
    performs the first step of the 7-step oidc handshake by
    registering the client and caching client credentials
    :param sso_oidc: the boto3 sso-oidc client
    :param bool force: specify this flag to bypass the cache
    :return: dict
    """
    cache = SsoTokenSaver('convocations')
    message = 'client credentials'
    expiration_key = 'clientSecretExpiresAt'
    if not force:
        creds = cache.load(expiration_key, message) or {}
        if creds:
            return creds
    notice('registering client')
    creds = sso_oidc.register_client(
        clientName='convocations',
        clientType='public',
    )
    notice_end()
    notice('caching response')
    cache.save(creds)
    notice_end()
    return creds


class SsoTokenSaver(SSOTokenLoader):
    """
    class that caches sso client credentials and device authorizations
    in the aws sso cache dir so that cached credentials play nicely with
    boto3 and botocore
    """
    def __init__(self, start_url):
        super().__init__()
        self.token_cache = JSONFileCache(cache_dir())
        self.start_url = start_url
        try:
            self.cache_key = self._generate_cache_key(start_url)
        except:
            self.cache_key = self._generate_cache_key(start_url, None)

    def save(self, response):
        """
        caches a token response or a client registration response
        :param dict response: the dict response returned from a call
            to `register_client` or `create_token`
        :return:
        """
        cache_key = self.cache_key
        self.token_cache[cache_key] = response

    def load(self, expiration_key, message='checking expiration'):
        """
        loads the cached response from the cache and checks its expiration.
        if the response is still valid (i.e., not expired), the response is
        returned.  Otherwise we return None
        :param expiration_key: the key in the response with expiration
        :param message: the notification message displayed as we check cache
        :return: the cached credentials, if valid, or none
        """
        try:
            notice(message)
            creds = self.token_cache[self.cache_key]
            expires_at = creds.get(expiration_key)
            if expires_at:
                now = int(time())
                delta = expires_at - now
                is_current = delta > 0
                if is_current:
                    notice_end(f'{delta / 3600:0.2f}h', 'green')
                    return creds
            notice_end('expired')
        except KeyError:
            notice_end('miss')
        return None


class SsoCredentialSaver:
    def __init__(self, start_url, role_name, account_id):
        self.token_cache = JSONFileCache(cache_dir())
        self.start_url = start_url
        self.role_name = role_name
        self.account_id = account_id
        self.cache_key = self.generate_cache_key()

    def generate_cache_key(self):
        args = {
            'startUrl': self.start_url,
            'roleName': self.role_name,
            'accountId': self.account_id,
        }
        # NOTE: It would be good to hoist this cache key construction logic
        # into the CachedCredentialFetcher class as we should be consistent.
        # Unfortunately, the current assume role fetchers that sub class don't
        # pass separators resulting in non-minified JSON. In the long term,
        # all fetchers should use the below caching scheme.
        args = json.dumps(args, sort_keys=True, separators=(',', ':'))
        h = sha1(args.encode('utf-8')).hexdigest().replace(':', '_')
        return h.replace(os.path.sep, '_')

    def _parse_timestamp(self, timestamp_ms):
        from pytz import utc
        # fromtimestamp expects seconds so: milliseconds / 1000 = seconds
        timestamp_seconds = timestamp_ms / 1000.0
        timestamp = datetime.fromtimestamp(timestamp_seconds, utc)
        return timestamp.strftime('%Y-%m-%dT%H:%M:%SZ')

    def save(self, credential_response):
        cache_key = self.cache_key
        credentials = credential_response['roleCredentials']
        self.token_cache[cache_key] = {
            'ProviderType': 'sso',
            'Credentials': {
                'AccessKeyId': credentials['accessKeyId'],
                'SecretAccessKey': credentials['secretAccessKey'],
                'SessionToken': credentials['sessionToken'],
                'Expiration': self._parse_timestamp(credentials['expiration']),
                'ExpirationTime': credentials['expiration'],
            }
        }

    def load(self):
        cache_key = self.cache_key
        try:
            notice('cached role creds')
            creds = self.token_cache[cache_key]
            creds = creds.get('Credentials') or {}
            expiration = creds.get('Expiration')
            expiration = parse_time(expiration)
            now = datetime.now(pytz.utc)
            delta = (expiration - now).total_seconds()
            is_current = delta > 0
            if is_current:
                notice_end(f'{delta / 3600:0.2f}h', 'green')
                return creds
            notice_end('expired')
            return None
        except KeyError:
            notice_end('not found')
            return None


def device_authorization(sso_oidc, client_creds, start_url, force=False):
    """
    performs steps 2-4 of the 7-step oidc handshake by performing the
    device authorization, redirecting the user to the sso sign-in page,
    and then exchanging the device auth for an access token once sign in
    is complete and authorized
    :param sso_oidc: the boto3 sso-oidc client
    :param client_creds: the client credentials response from `register_client`
    :param start_url: the sso start url specified in the profile or the context
    :param force: specify this flag to bypass use of cached credentials
    :return:
    """
    cache = SsoTokenSaver(start_url)
    message = 'cached device authorization'
    expiration_key = 'accessTokenExpiration'
    if not force:
        creds = cache.load(expiration_key, message) or {}
        if creds:
            return False, creds
    notice('device authorization')
    response = sso_oidc.start_device_authorization(
        clientId=client_creds['clientId'],
        clientSecret=client_creds['clientSecret'],
        startUrl=start_url,
    )
    device_code = response['deviceCode']
    url = response['verificationUriComplete']
    expires_in = response['expiresIn']
    interval = response['interval']
    interval = max(interval, 7)
    notice_end(f'{url} / {expires_in} / {interval}')
    webbrowser.open(url, autoraise=True)
    for n in range(1, expires_in // interval + 1):
        notice(f'poll #{n}')
        sleep(interval)
        try:
            creds = sso_oidc.create_token(
                grantType='urn:ietf:params:oauth:grant-type:device_code',
                deviceCode=device_code,
                clientId=client_creds['clientId'],
                clientSecret=client_creds['clientSecret'],
            )
            creds['accessTokenExpiration'] = creds['expiresIn'] + time()
            creds['expiresAt'] = creds['accessTokenExpiration']
            cache.save(creds)
            notice_end()
            return True, creds
        except sso_oidc.exceptions.AuthorizationPendingException:
            notice_end('wait')
    return False, None


def get_role(sso, role, account_id, device_auth):
    """
    performs steps 5 of the 7-step oidc handshake by performing the
    role selection automatically. we will choose the first role available
    alphabetically if no role is specified
    :param sso: the boto3 sso client
    :param device_auth: the device auth response from `create_token`
    :param role: the role to choose, can be None to just pick the first role
    :param account_id: the account id we are signing into
    :return: the role we should use
    """
    notice('listing account roles')
    access_token = device_auth['accessToken']
    try:
        account_roles = sso.list_account_roles(
            accessToken=access_token,
            accountId=account_id,
        )
    except sso.exceptions.UnauthorizedException:
        notice_end('refresh token')
        return None
    roles = account_roles['roleList']
    all_roles = [ x['roleName'] for x in roles ]
    if 'AWSOrganizationsFullAccess' in all_roles:
        all_roles.remove('AWSOrganizationsFullAccess')
    all_roles.sort(key=lambda lx: lx.lower())
    if role and role not in all_roles:
        notice_end(f'could not find {role}.  Please try: {", ".join(all_roles)}')
        return None
    role = role or all_roles[0]
    notice_end(f'{role} / {", ".join(all_roles)}')
    return role


def get_role_creds(sso, device_auth, account_id, start_url, role, force=False):
    """
    performs step 6 of the 7-step oidc handshake which is exchanging the
    access token for the role credentials we can use.

    the response will be formatted as:

    ```
    {
        'AccessKeyId': 'AK.....',
        'SecretAccessKey': '01...',
        'SessionToken': '....',
        'Expiration': '2022-09-10T10:00:00'
    }
    ```

    :param sso: the boto3 sso client
    :param device_auth: the response from a previous call to `create_token`
    :param account_id: the account id we are signing into
    :param start_url: the start_url that initiates the oidc process
    :param role: the role that we are assuming
    :param force: specify this flag to bypass the cache
    :return: the cached credential formatted response to `get_role_credentials`
    """
    cache = SsoCredentialSaver(start_url, role, account_id)
    if not force:
        creds = cache.load() or {}
        if creds:
            return creds
    notice(f'getting credentials for {role}')
    cache = SsoCredentialSaver(start_url, role, account_id)
    access_token = device_auth['accessToken']
    creds = sso.get_role_credentials(
        roleName=role,
        accountId=account_id,
        accessToken=access_token,
    )
    cache.save(creds)
    notice_end()
    return cache.load()


def unpack_sso_config(ctx: Context, sso_config: dict, role: str) -> Tuple[str, str, str, str]:
    """
    values in the context are prioritized over values in the aws config
    """
    account_id = get_context_value(ctx, 'aws.account_id')
    account_id = account_id or get_context_value(ctx, 'aws.sso_account_id')
    account_id = account_id or sso_config.get('sso_account_id')
    account_id = f'{account_id}'
    sso_region = get_context_value(ctx, 'aws.sso_region')
    sso_region = sso_region or sso_config.get('sso_region')
    start_url = get_context_value(ctx, 'aws.start_url')
    start_url = start_url or sso_config.get('sso_start_url')
    role = role or get_context_value(ctx, 'aws.role')
    role = role or sso_config.get('sso_role_name')
    return account_id, sso_region, start_url, role


def make_sso_session(
        ctx: Optional[Context] = None,
        role: Optional[str] = None,
        session: Optional[boto3.session.Session] = None,
        profile: Optional[str] = None,
        force_sso: Optional[bool] = False,
        sso_config: Optional[dict] = None) -> Optional[boto3.session.Session]:
    """
    performs the 7-step oidc handshake to create an sso session with aws.
    see explanation here https://stackoverflow.com/a/71850591/6043170
    """
    from botocore.session import Session as CoreSession
    from boto3 import Session
    if session:
        return session
    if profile and not sso_config:
        core_session = CoreSession(profile=profile)
        profiles = core_session.full_config.get('profiles', {})
        sso_config = profiles.get(profile, {})
    sso_config = sso_config or {}
    t = unpack_sso_config(ctx, sso_config, role)
    account_id, sso_region, start_url, role = t
    region = get_context_value(ctx, 'aws.region')
    profile = profile or get_context_value(ctx, 'aws.profile')

    session = Session(region_name=sso_region)
    sso_oidc = session.client('sso-oidc')
    client_creds = client_credentials(sso_oidc, force_sso)
    reauth, device_auth = device_authorization(
        sso_oidc, client_creds, start_url, force_sso)
    if not device_auth:
        return None
    sso = session.client('sso')
    role = get_role(sso, role, account_id, device_auth)
    if not role:
        return None
    role_creds = get_role_creds(
        sso, device_auth,
        account_id, start_url, role,
        force_sso or reauth)
    secret_access_key = role_creds['SecretAccessKey']
    access_key_id = role_creds['AccessKeyId']
    session = Session(
        region_name=region,
        profile_name=profile,
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        aws_session_token=role_creds['SessionToken'],
    )
    get_account_id(session)
    get_alias(session)
    return session


def get_account_id(session: boto3.session.Session) -> Optional[str]:
    """
    gets the account id we are signed into and caches the result
    as an attribute on the `session` object
    :param session: a boto3.session.Session object
    :return str: the aws account id (usually a 12-digit number)
    """
    account_id = getattr(session, 'account_id', None)
    if not account_id:
        sts = session.client('sts')
        notice('account id')
        data = sts.get_caller_identity()
        account_id = session.account_id = data['Account']
        notice_end(data['Account'], 'yellow', attrs=[ 'bold' ])
        notice('username')
        session.whoami = data['Arn'].rsplit('/', 1)[-1]
        notice_end(session.whoami, 'blue', attrs=[ 'underline' ])
    return account_id


def get_alias(session: boto3.session.Session) -> Optional[str]:
    """
    gets the iam account alias of the account we are signed into and
    caches the result as an attribute on the `session` object
    :param session: a boto3.session.Session object
    :return str: the human understandable iam account alias
    """
    alias = getattr(session, 'alias', None)
    if not alias:
        notice('alias')
        iam = session.client('iam')
        aliases = iam.list_account_aliases()
        aliases = aliases['AccountAliases']
        alias = session.alias = aliases[0] if aliases else 'unknown'
        notice_end(f'{session.alias}', 'yellow', attrs=[ 'bold' ])
    return alias


def make_session(profile, session=None, region=None):
    """
    creates an aws session using the provided profile and region
    :param str profile: the name of the aws profile
    :param boto3.session.Session session: an existing session if one exists
    :param str region: the aws region -- e.g., us-east-2
    :rtype: boto3.session.Session
    """
    if session:
        return session
    notice(f'checking profile {profile}')
    session = boto3.Session()
    if profile not in session.available_profiles:
        notice_end('not found')
        return None
    notice_end()
    session = boto3.Session(profile_name=profile, region_name=region)
    get_account_id(session)
    get_alias(session)
    return session


class AwsTask(Task):
    def __call__(self, *args, **kwargs):
        profile = kwargs.get('profile')
        ctx = args[0]
        session = kwargs.get('session')
        has_context = isinstance(ctx, Context)
        if has_context:
            session = session or getattr(ctx, 'context', None)
        prefix = kwargs.get('prefix')
        force_sso = kwargs.pop('force_sso', False)
        if profile is None and has_context:
            profile = get_context_value(ctx, 'aws.profile')
        if prefix is None and has_context:
            prefix = get_context_value(ctx, 'aws.prefix')
        notice('is sso?')
        sso_config = None
        is_sso = get_context_value(ctx, 'aws.start_url')
        is_sso = bool(is_sso)
        if not is_sso:
            from botocore.session import Session as CoreSession
            if profile:
                core_session = CoreSession(profile=profile)
                profiles = core_session.full_config.get('profiles', {})
                sso_config = profiles.get(profile, {})
                is_sso = sso_config.get('sso_start_url')
                is_sso = is_sso and sso_config.get('sso_account_id')
                is_sso = is_sso and sso_config.get('sso_region')
        notice_end(is_sso)
        if is_sso:
            role = kwargs.get('role') or get_context_value(ctx, 'aws.role_name')
            session = make_sso_session(
                ctx, role, session, profile, force_sso,
                sso_config)
        else:
            session = make_session(profile, session)
        if has_context:
            setattr(ctx, 'session', session)
        kwargs['session'] = session
        kwargs['prefix'] = prefix
        return super().__call__(*args, **kwargs)


def get_tag(p, key):
    try:
        tags = p.get('Tags', [])
    except AttributeError:
        tags = p.tags or []
    key = key.lower()
    for x in tags:
        if x['Key'].lower() == key:
            return x['Value']
    return ''


def name_tag(p: Any, lower=True) -> str:
    """
    searches a dict or an object for tags with the value name and returns the
    first such tag value
    """
    name = get_tag(p, 'Name')
    if lower:
        name = name.lower()
    return name


def yielder(client, method: str, session=None, **kwargs) -> List[Dict]:
    if not session:
        return
    if isinstance(client, str):
        client = session.client(client)
    p = client.get_paginator(method)
    for page in p.paginate(**kwargs):
        for key, value in page.items():
            if key != 'NextMarker' and isinstance(value, list):
                yield from value


def get_path(data, path, default=None):
    """
    allows access to a dict by the dotted path
    """
    keys = path.split('.')
    value = data
    for x in keys:
        if isinstance(value, Mapping) and x in value:
            value = value[x]
        else:
            return default
    return value


def post_to_slack(ctx, session, room, message):
    session = session or boto3.DEFAULT_SESSION
    token = getattr(ctx, f'{room}_token', None)
    if not token:
        notice(f'getting {room} token')
        ssm = session.client('ssm')
        try:
            response = ssm.get_parameter(Name=f'/dev/slack/{room}')
            token = response['Parameter']['Value']
            notice_end()
        except ClientError as ex:
            notice_end(ex.response['Error']['Code'])
            return False
    message = f'*{getuser()}* {message}'
    url = f'https://hooks.slack.com/services/{token}'
    return requests.post(url, json={ 'text': message }, timeout=5)



from raft import task
from raft.collection import Collection


def render(env, source, destination, context=None):
    """
    :param jinja2.Environment env:
    :param str source:
    :param str destination:
    :param dict[Any, Any] context:
    """
    t = env.get_template(source)
    context = context or {}
    body = t.render(**context)
    with open(destination, 'w') as f:
        f.write(body)


@task(help=dict(
    name='the name of the project',
    path='the parent directory in which {name} will be created',
    source_dir='the name of the source directory within our project, for '
               'most projects, leave this blank so that the value defaults '
               'to {name}',
    skip_setup='if this flag is specified, we will not create a virtual '
               'environment and install base requirements as part of '
               'kicking off the new directory',
))
def template(ctx, name, path='.', source_dir=None, skip_setup=False):
    """
    creates a template for a pypi project using our standard defaults
    """
    from ..base.projects import new_project
    boo = not skip_setup
    templates = [
        '.bumpversion.cfg',
        '.coveragerc',
        '.gitignore',
        'dev.txt',
        'Makefile',
        'MANIFEST.in',
        'pytest.ini',
        'README.md',
        'requirements.txt',
        'setup.py',
        'sesame_meow_cat/__init__.py',
        'sesame_meow_cat/version.txt',
        'test/__init__.py',
        'test/conftest.py',
    ]
    new_project(
        ctx, name, path, source_dir, boo,
        templates, package='convocations.pypi')
    # from jinja2.environment import Environment
    # from jinja2.loaders import PackageLoader
    # print(f'checking for existence of {path}')
    # if not os.path.exists(path):
    #     os.makedirs(path, exist_ok=True)
    # print(f'cding to {path}')
    # os.chdir(path)
    # os.makedirs(name, exist_ok=True)
    # source_dir = source_dir or name
    # root_dir = name
    # base_dir = os.path.join(root_dir, source_dir)
    # test_dir = os.path.join(root_dir, 'test')
    # print(f'base_dir: [{base_dir}]')
    # print(f'test_dir: [{test_dir}]')
    # os.makedirs(base_dir, exist_ok=True)
    # os.makedirs(test_dir, exist_ok=True)
    # loader = PackageLoader('convocations.pypi', 'templates')
    # env = Environment(loader=loader)
    # context = dict(name=name)
    # files = [
    #     ('.bumpversion.cfg.j2', root_dir, '.bumpversion.cfg'),
    #     ('.coveragerc.j2', root_dir, '.coveragerc'),
    #     ('.gitignore.j2', root_dir, '.gitignore'),
    #     ('empty_init.py.j2', base_dir, '__init__.py'),
    #     ('version.txt.j2', base_dir, 'version.txt'),
    #     ('makefile.j2', root_dir, 'Makefile'),
    #     ('dev.txt.j2', root_dir, 'dev.txt'),
    #     ('requirements.txt.j2', root_dir, 'requirements.txt'),
    #     ('MANIFEST.in.j2', root_dir, 'MANIFEST.in'),
    #     ('pytest.ini.j2', root_dir, 'pytest.ini'),
    #     ('conftest.py.j2', root_dir, 'test', 'conftest.py'),
    #     ('empty_init.py.j2', root_dir, 'test', '__init__.py'),
    #     ('README.md.j2', root_dir, 'README.md'),
    #     ('setup.py.j2', root_dir, 'setup.py'),
    # ]
    # for source, *paths in files:
    #     filename = os.path.join(*paths)
    #     if not os.path.exists(filename):
    #         print(f'rendering {source} => {filename}')
    #         render(env, source, filename, context)
    # with ctx.cd(root_dir):
    #     ctx.run('make setup')


ns = Collection(template)



from dataclasses import dataclass
from typing import List

from raft import task


@dataclass
class Recordset:
    hosted_zone_name: str = None
    type_: str = None
    name: str = None
    ttl: str = None
    resource_records: List[str] = None

    @property
    def normalized_name(self):
        name = self.name
        if not name.endswith('.'):
            name = f'{name}.'
        return name

    @property
    def normalized_hosted_zone_name(self):
        name = self.hosted_zone_name
        if not name.endswith('.'):
            name = f'{name}.'
        return name

    @property
    def logical_name(self):
        name = self.normalized_name
        pieces = name.split('.')[:-3]
        if pieces:
            pieces = [ x.replace('_', 'underscore').title() for x in pieces ]
            rg = []
            for x in pieces:
                x = ''.join([ lx.title() for lx in x.split('-') ])
                rg.append(x)
            pieces = rg
            lname = ''.join(pieces)
        else:
            type_ = self.type_.title()
            lname = f'Apex{type_}'
        return lname

    @property
    def logical_name_with_type(self):
        name = self.normalized_name
        pieces = name.split('.')[:-3]
        if pieces:
            pieces = [ x.replace('_', 'underscore').title() for x in pieces ]
            rg = []
            for x in pieces:
                x = ''.join([ lx.title() for lx in x.split('-') ])
                rg.append(x)
            pieces = rg
            pieces.append(self.type_.title())
            lname = ''.join(pieces)
        else:
            type_ = self.type_.title()
            lname = f'Apex{type_}'
        return lname

    def cfn(self):
        properties = dict(
            HostedZoneName=self.normalized_hosted_zone_name,
            Name=self.normalized_name,
            TTL=self.ttl,
            Type=self.type_,
            ResourceRecords=self.resource_records,
        )
        data = {
            self.logical_name: {
                'Type': 'AWS::Route53::RecordSet',
                'Properties': properties,
            },
        }
        return data


def parse_zone_line(st):
    pieces = st.split(' ', 4)
    name = pieces[0]
    if not name.endswith('.'):
        name = f'{name}.'
    hosted_zone_name = name.split('.')[-3:]
    hosted_zone_name = '.'.join(hosted_zone_name)
    if '"' in pieces[-1]:
        values = pieces[-1].split('"')[1::2]
        values = [ f'"{x}"' for x in values ]
    else:
        values = [ pieces[-1] ]
    return Recordset(
        hosted_zone_name=hosted_zone_name,
        name=name,
        ttl=pieces[1],
        type_=pieces[3],
        resource_records=values,
    )


@task
def parse_zone_file(ctx, filename):
    """
    parses a zone file and reformats as a cloudformation template
    """
    from convocations.base.utils import dump_yaml
    with open(filename, 'r') as f:
        raw_lines = f.readlines()
    lines = [ x.strip() for x in raw_lines if not x.startswith(';') ]
    resources = {}
    recordsets = {}
    for x in lines:
        recordset = parse_zone_line(x)
        names = [ recordset.logical_name, recordset.logical_name_with_type ]
        for logical_name in names:
            if logical_name not in recordsets:
                recordsets[logical_name] = recordset
                break
            r = recordsets[logical_name]
            if r.type_ == recordset.type_:
                r.resource_records += recordset.resource_records
                break

    for x in recordsets.values():
        cfn = x.cfn()
        resources.update(cfn)
    cfn = dict(AWSTemplateFormatVersion='2010-09-09', Resources=resources)
    dump_yaml(cfn)




from raft import task
from convocations.base.utils import notice, notice_end
from convocations.base.utils import print_table, dump_yaml
from ..base import AwsTask, yielder


def get_service(ctx, name, cluster, session):
    """
    gets a service by name, case-insensitve
    """
    ecs = session.client('ecs')
    notice(f'looking for ecs service [{name}]')
    found_services = ecs_services(ctx, name, cluster, session=session, quiet=True)
    if not found_services:
        notice_end('nothing found')
        return None
    notice_end(len(found_services))
    name = name.lower()
    cluster, service = None, None
    if len(found_services) > 1:
        for x, y in found_services:
            if y.lower() == name:
                cluster, service = x, y
                break
    if not cluster:
        cluster, service = found_services[0]
    response = ecs.describe_services(cluster=cluster, services=[ service ])
    services = response.get('services')
    if services:
        notice_end()
        return services[0]
    notice_end('not found')
    return None


def get_cluster(name, session):
    notice(f'looking up {name}')
    ecs = session.client('ecs')
    response = ecs.describe_clusters(clusters=[ name ], include=[
        'ATTACHMENTS', 'CONFIGURATIONS', 'SETTINGS', 'TAGS',
    ])
    clusters = response['clusters']
    if not clusters:
        notice_end('not found')
        return None
    return clusters[0]


def get_task_definition(name, session):
    """
    gets a task definition given the  name of the task definition.

    if just the family name for the ecs task is provided, we will
    get the definition for the highest-numbered, active task definition.
    to get a particular task definition, send in the `family_name:version`
    formatted name.
    """
    ecs = session.client('ecs')
    notice(f'describing [{name}]')
    response = ecs.describe_task_definition(taskDefinition=name)
    t_def = response.get('taskDefinition')
    arn = t_def['taskDefinitionArn']
    notice_end(arn.rsplit('/', 1)[-1])
    return t_def


@task(klass=AwsTask)
def service_to_yaml(ctx, name=None, cluster=None, session=None, quiet=False, **kwargs):
    """
    generates the cloudformation yaml for an ecs service
    """
    from .import_resource_to_cfn import to_pascal_case
    service = get_service(ctx, name, cluster, session)
    if not service:
        return None
    logical_name = ''.join(map(lambda lx: lx.title(), name.split('-')))
    load_balancers = to_pascal_case(service['loadBalancers'])
    service_registries = to_pascal_case(service.get('serviceRegistries'))
    capacity_provider_strategy = to_pascal_case(service.get('capacityProviderStrategy'))
    properties = dict(
        CapacityProviderStrategy=capacity_provider_strategy,
        Cluster=service['clusterArn'].split('/', 1)[-1],
        DeploymentConfiguration=to_pascal_case(service.get('deploymentConfiguration')),
        DeploymentController=None,
        DesiredCount=service['desiredCount'],
        EnableECSManagedTags=service.get('enableECSManagedTags'),
        EnableExecuteCommand=service.get('enableExecuteCommand'),
        HealthCheckGracePeriodSeconds=service.get('healthCheckGracePeriodSeconds'),
        LaunchType=service['launchType'],
        LoadBalancers=load_balancers or None,
        NetworkConfiguration=to_pascal_case(service.get('networkConfiguration')) or None,
        PlacementConstraints=to_pascal_case(service.get('placementConstraints')) or None,
        PlacementStrategies=to_pascal_case(service.get('placementStrategy')) or None,
        PlatformVersion=service.get('platformVersion'),
        PropagateTags=service.get('propagateTags'),
        Role=service.get('roleArn'),
        SchedulingStrategy=service.get('schedulingStrategy') or None,
        ServiceConnectConfiguration=None,
        ServiceName=service['serviceName'],
        ServiceRegistries=service_registries or None,
        Tags=to_pascal_case(service.get('tags')) or None,
        TaskDefinition=service['taskDefinition'].split('/', 1)[-1],
    )
    properties = { k: v for k, v in properties.items() if v is not None }
    data = {
        logical_name: dict(
            Type='AWS::ECS::Service',
            DeletionPolicy='Retain',
            Properties=properties,
        ),
    }
    if not quiet:
        dump_yaml(data)
    return data


@task(klass=AwsTask)
def cluster_to_yaml(ctx, name=None, session=None, quiet=False, **kwargs):
    """
    generates the cloudformation yaml for an ecs service
    """
    from .import_resource_to_cfn import to_pascal_case
    cluster = get_cluster(name, session)
    if not cluster:
        return None
    logical_name = ''.join(map(lambda lx: lx.title(), name.split('-')))
    properties = dict(
        CapacityProviders=cluster.get('capacityProviders') or None,
        ClusterName=cluster['clusterName'],
        ClusterSettings=to_pascal_case(cluster.get('settings')) or None,
        Configuration=to_pascal_case(cluster.get('configuraiton')) or None,
        DefaultCapacityProviderStrategy=to_pascal_case(cluster.get('defaultCapacityProviderStrategy')) or None,
        ServiceConnectDefaults=to_pascal_case(cluster.get('serviceConnectDefaults')) or None,
        Tags=to_pascal_case(cluster.get('tags')) or None,
    )

    properties = { k: v for k, v in properties.items() if v is not None }
    data = {
        logical_name: dict(
            Type='AWS::ECS::Cluster',
            DeletionPolicy='Retain',
            Properties=properties,
        ),
    }
    if not quiet:
        dump_yaml(data)
    return data


@task(klass=AwsTask)
def task_definition_to_yaml(ctx, name=None, session=None, quiet=False, **kwargs):
    """
    generates the cloudformation yaml for an ecs task definition
    """
    from .import_resource_to_cfn import to_pascal_case
    t_def = get_task_definition(name, session)
    if not t_def:
        return None
    logical_name = ''.join(map(lambda lx: lx.title(), name.split('-')))
    properties = dict(
        ContainerDefinitions=to_pascal_case(t_def['containerDefinitions'], exclude=['Options']),
        Cpu=t_def.get('cpu'),
        EphemeralStorage=to_pascal_case(t_def.get('ephemeralStorage')) or None,
        ExecutionRoleArn=t_def.get('executionRoleArn'),
        Family=t_def.get('family'),
        InferenceAccelerators=to_pascal_case(t_def.get('inferenceAccelerators')) or None,
        IpcMode=t_def.get('ipcMode'),
        Memory=t_def.get('memory'),
        NetworkMode=t_def.get('networkMode'),
        PidMode=t_def.get('pidModer'),
        PlacementConstraints=to_pascal_case(t_def.get('placementsConstraints')) or None,
        ProxyConfiguration=to_pascal_case(t_def.get('proxyConfiguration')) or None,
        RequiresCompatibilities=t_def.get('requiresCompatabilities') or None,
        RuntimePlatform=to_pascal_case(t_def.get('runtimePlatform')) or None,
        Tags=to_pascal_case(t_def.get('tags')) or None,
        TaskRoleArn=t_def.get('taskRoleArn'),
        Volumes=to_pascal_case(t_def.get('volumes')) or None,
    )

    properties = { k: v for k, v in properties.items() if v is not None }
    data = {
        logical_name: dict(
            Type='AWS::ECS::TaskDefinition',
            DeletionPolicy='Retain',
            Properties=properties,
        ),
    }
    if not quiet:
        dump_yaml(data)
    return data


@task(klass=AwsTask)
def ecs_clusters(ctx, name=None, session=None, quiet=False, **kwargs):
    """
    lists all ecs clusters that contain `name` in their name or name tag
    """
    rows = []
    clusters = []
    if not quiet:
        notice('listing clusters')
    for x in yielder('ecs', 'list_clusters', session=session):
        short_name = x.rsplit('/', 1)[-1]
        if not name or name.lower() in short_name.lower():
            clusters.append(short_name)
            rows.append([ short_name ])
    if not quiet:
        notice_end(len(clusters))
    if not quiet:
        rows.sort(key=lambda lx: lx[0])
        print()
        print_table([ 'name' ], rows)
        print()
    clusters.sort()
    return clusters


@task(klass=AwsTask)
def ecs_services(ctx, name=None, cluster=None, session=None, quiet=False, **kwargs):
    """
    lists all ecs services that contain `name` in their name or name tag

    * searches across all clusters
    """
    rows = []
    clusters = ecs_clusters(ctx, cluster, session=session, quiet=quiet)
    if not clusters:
        return []
    services = []
    for k in clusters:
        if not quiet:
            notice(f'listing cluster {k}')
        params = dict(cluster=k)
        service_rows = []
        for x in yielder('ecs', 'list_services', **params, session=session):
            short_name = x.rsplit('/', 1)[-1]
            if not name or name.lower() in short_name.lower():
                service_rows.append([ short_name, k ])
                services.append([ k, short_name ])
        rows += service_rows
        if not quiet:
            notice_end(len(service_rows))
    if not quiet:
        print()
        print_table([ 'name', 'cluster' ], rows)
        print()
    return services



from configparser import ConfigParser
from raft import task
from ..base.utils import notice_end, notice


@task
def generate_fernet_key(ctx, filename=None):
    """
    generates a fernet key for the airflow.cfg file
    if no fernet key exists within the current configuration.  If a fernet
    key exists within the current configuration, we simply ignore.
    """
    filename = filename or ctx.airflow.config
    p = ConfigParser()
    p.read(filename)
    fernet_key = p['core'].get('fernet_key')
    if not fernet_key:
        from cryptography.fernet import Fernet
        fernet_key = Fernet.generate_key().decode()
        p['core']['fernet_key'] = fernet_key
        with open(filename, 'w') as f:
            p.write(f)


@task
def setup_local_connections(ctx, key=None):
    """
    installs connections on your local airflow
    """
    import requests
    conf = ctx.airflow
    session = requests.Session()
    session.auth = (conf.username, conf.password)
    base_url = f'{conf.base_url}/api/v1/connections'
    response = session.get(base_url)
    data = response.json()
    connections = conf.connections
    existing_connections = data['connections']
    connection_ids = { x['connection_id'] for x in existing_connections }
    for connection_id in connections.keys():
        if key and key != connection_id:
            continue
        found = connection_id in connection_ids
        if found:
            fn = session.patch
            url = f'{base_url}/{connection_id}'
            notice(f'patching {connection_id}')
        else:
            fn = session.post
            url = base_url
            notice(f'creating {connection_id}')
        x = connections[connection_id]
        body = dict(
            conn_type=x.get('conn_type') or '',
            host=x.get('host') or '',
            login=x.get('login') or '',
            password=x.get('password') or '',
            schema=x.get('schema') or '',
            port=x.get('port') or None,
            extra=x.get('extra') or ''
        )
        response = fn(url, json=body)
        notice_end(response.text)


@task
def setup_local_variables(ctx):
    """
    installs variables on your local airflow running in docker compose
    """
    import requests
    conf = ctx.airflow
    variables = conf.variables
    if not variables:
        print('nothing to be done')
        return
    session = requests.Session()
    session.auth = (conf.username, conf.password)
    base_url = f'{conf.base_url}/api/v1/variables'
    response = session.get(base_url)
    data = response.json()
    existing_variables = { x['key'] for x in data['variables'] }
    for key, value in variables.items():
        found = key in existing_variables
        if found:
            fn = session.patch
            url = f'{base_url}/{key}'
            notice(f'patching {key}')
        else:
            fn = session.post
            url = base_url
            notice(f'creating {key}')
        body = dict(key=key, value=value)
        fn(url, json=body)
        notice_end()


@task
def clear_queue(ctx, filename):
    """
    clears the celery queue to prevent weird errors
    `TypeError: unsupported operand type(s) for +: 'float' and 'str'`
    """
    from redis import Redis
    filename = filename or ctx.airflow.config
    p = ConfigParser()
    p.read(filename)
    notice('looking up broker url')
    broker_url = p['celery'].get('broker_url')
    if not broker_url:
        cmd = p['celery'].get('broker_url_cmd')
        if cmd:
            result = ctx.run(cmd)
            broker_url = result.stdout.strip()
    if not broker_url:
        notice_end('not found')
        return
    notice_end()
    notice('clearing default queue')
    client = Redis(broker_url)
    client.delete('default')
    notice_end()



from getpass import getpass
from raft.tasks import task

from ...base.utils import notice, notice_end
from ..base import AwsTask


__all__ = [
    'add_user',
    'set_password',
]


@task(klass=AwsTask)
def add_user(ctx, username, group, session=None, **kwargs):
    """
    Adds a user to the specified group
    """
    notice(f'creating user {username}')
    iam = session.resource('iam')
    user = iam.User(username)
    user_created = False
    try:
        user.load()
        notice_end('found')
    except iam.meta.client.exceptions.NoSuchEntityException:
        user.create()
        user.reload()
        notice_end('created')

    notice(f'has group {group}?')
    has_group = False
    for x in user.groups.all():
        if x.name.lower() == group.lower():
            has_group = True
            break
    notice_end(f'{has_group}')

    if not has_group:
        notice(f'adding {username} to {group}')
        user.add_group(GroupName=group)
        notice_end()

    if user_created:
        set_password(ctx, username, session=session)


@task(klass=AwsTask)
def set_password(ctx, username, session=None):
    """
    sets the password for the provided username
    """
    iam = session.resource('iam')
    user = iam.User(username)
    notice(f'loading user {username}')
    try:
        user.load()
        notice_end('found')
    except iam.meta.client.exceptions.NoSuchEntityException:
        notice_end('not found')
        return

    notice('profile exists?')
    profile = user.LoginProfile()
    try:
        profile.load()
        fn = profile.update
        notice_end('yes')
    except iam.meta.client.exceptions.NoSuchEntityException:
        fn = profile.create
        notice_end('no')
    password = getpass('Please enter your password: ')
    confirm = getpass('Confirm                   : ')
    if password == confirm:
        notice('updating password')
        fn(Password=password)
        notice_end()
    else:
        notice_end('passwords do not match')




from raft import task
from ..base import AwsTask
from ..ec2 import instance_by_name
from ..ec2 import get_tag


def get_instance(ctx, name, instance_id, session, instance=None):
    if not instance:
        if name and not instance_id:
            instance = instance_by_name(ctx, name, session, None)
        elif instance_id:
            ec2 = session.resource('ec2')
            instance = ec2.Instance(instance_id)
    if not instance:
        return None, None
    name = get_tag(instance, 'name')
    return name, instance


@task(klass=AwsTask)
def instance_to_yaml(ctx, name=None, instance_id=None, instance=None,
                     session=None, quiet=False, **kwargs):
    """
    generates the cloudformation yaml for an instance
    """
    name, instance = get_instance(ctx, name, instance_id, session, instance)
    logical_name = ''.join(map(lambda lx: lx.title(), name.split('-')))
    volumes = [dict(
        Device=x['DeviceName'],
        VolumeId=x['Ebs']['VolumeId'],
    ) for x in instance.block_device_mappings]
    network_interfaces = [dict(
        DeviceIndex=n,
        NetworkInterfaceId=x.network_interface_id,
    ) for n, x in enumerate(instance.network_interfaces, 1)]
    security_group_ids = [ x['GroupId'] for x in instance.security_groups ]
    profile_name = (instance.iam_instance_profile or {}).get('Arn', '')
    profile_name = profile_name.rsplit('/', 1)[-1]
    properties = dict(
        EbsOptimized=instance.ebs_optimized,
        IamInstanceProfile=profile_name or None,
        ImageId=instance.image_id,
        InstanceType=instance.instance_type,
        KeyName=instance.key_name,
        NetworkInterfaces=network_interfaces,
        SecurityGroupIds=security_group_ids,
        SourceDestCheck=instance.source_dest_check,
        SubnetId=instance.subnet_id,
        Tags=instance.tags,
        Volumes=volumes,
    )
    properties = { k: v for k, v in properties.items() if v is not None }
    data = {
        logical_name: dict(
            Type='AWS::EC2::Instance',
            DeletionPolicy='Retain',
            Properties=properties,
        ),
    }
    if not quiet:
        from convocations.base.utils import dump_yaml
        dump_yaml(data)
    return data


@task(klass=AwsTask)
def import_instance(ctx, name=None, instance_id=None, filename=None,
                    skip_review=False, session=None, prefix=None, **kwargs):
    """
    creates a yaml template for an instance and imports it into cloudformation

    you can provide either the name of the instance or its instance id.
    if you specify both, the name will be ignored.
    if you specify the name, the instance id will be looked up based
      on a case-insensitive match with the name
    the filename is the name of the file and stack we will create

    if skip_review is set, we will immediately try to execute the change set
      in cloudformation
    """
    from .import_resource_to_cfn import import_resource_to_cfn
    name, instance = get_instance(ctx, name, instance_id, session)
    if not instance:
        return
    data = instance_to_yaml(
        ctx, None, None, instance,
        session=session, quiet=True)
    resources_to_import = [dict(
        ResourceType='AWS::EC2::Instance',
        LogicalResourceId=list(data.keys())[0],
        ResourceIdentifier=dict(InstanceId=instance.id),
    )]
    import_resource_to_cfn(
        data, resources_to_import,
        filename, name, skip_review, session, prefix)



import time

from raft import task
from ...base.utils import notice, notice_end, confirm
from ...base.utils import print_table
from ..ec2 import get_tag
from ..base import AwsTask, name_tag


@task(klass=AwsTask)
def transit_routes(ctx, session=None, profile=None, **kwargs):
    """
    lists all transit gateway routes
    """
    notice('getting transit gateway route tables')
    ec2 = session.client('ec2')
    gateways = ec2.describe_transit_gateway_route_tables()
    table_ids = [
        x['TransitGatewayRouteTableId']
        for x in gateways['TransitGatewayRouteTables']
    ]
    notice_end()
    print_table([ 'table id' ], [ [ x ] for x in table_ids ])
    rows = []
    header = [ 'table_id', 'cidr', 'type', 'resource_id', 'account', 'name', ]
    attachments = {}
    for table_id in table_ids:
        notice(f'getting routes for {table_id}')
        routes = ec2.search_transit_gateway_routes(
            TransitGatewayRouteTableId=table_id,
            Filters=[{
                'Name': 'state',
                'Values': [ 'active', ],
            }]
        )
        for route in routes['Routes']:
            attachment = route['TransitGatewayAttachments'][0]
            attachment_id = attachment['TransitGatewayAttachmentId']
            row = [
                table_id,
                route['DestinationCidrBlock'],
                attachment['ResourceType'],
                attachment['ResourceId'],
            ]
            detail = attachments.get(attachment_id)
            if not detail:
                detail = ec2.describe_transit_gateway_attachments(
                    TransitGatewayAttachmentIds=[
                        attachment_id,
                    ],
                )
                detail = detail['TransitGatewayAttachments'][0]
            account = detail['ResourceOwnerId']
            name = get_tag(detail, 'Name')
            row.append(account)
            row.append(name or '')
            rows.append(row)
        notice_end()
    print()
    print_table(header, rows)


@task(klass=AwsTask)
def accept_resource_shares(ctx, session=None, profile=None, **kwargs):
    """
    accepts any pending resource shares
    """
    notice('getting invitations')
    ram = session.client('ram')
    invitations = ram.get_resource_share_invitations()
    invitations = invitations['resourceShareInvitations']
    invitations = [ x for x in invitations if x['status'] == 'PENDING' ]
    notice_end(f'{len(invitations)} pending invitations')
    for x in invitations:
        arn = x['resourceShareInvitationArn']
        account_id = x['senderAccountId']
        notice(f'accepting {account_id} / {arn}')
        ram.accept_resource_share_invitation(resourceShareInvitationArn=arn)
        notice_end()


@task(klass=AwsTask)
def accept_pending_tgw_attachments(ctx, session=None, profile=None, **kwargs):
    """
    accepts any pending resource shares
    """
    notice('getting pending attachments')
    ec2 = session.client('ec2')
    attachments = ec2.describe_transit_gateway_attachments(Filters=[dict(
        Name='state',
        Values=[
            'pending',
            'pendingAcceptance',
        ]
    )])
    attachments = list(attachments['TransitGatewayAttachments'])
    notice_end(f'{len(attachments)} pending')
    for x in attachments:
        n_id = x['TransitGatewayAttachmentId']
        owner = x['ResourceOwnerId']
        resource_type = x['ResourceType']
        if confirm(f'accept {owner} / {n_id}'):
            notice('accepting')
            if resource_type == 'vpc':
                ec2.accept_transit_gateway_vpc_attachment(
                    TransitGatewayAttachmentId=n_id
                )
            elif resource_type == 'peering':
                ec2.accept_transit_gateway_peering_attachment(
                    TransitGatewayAttachmentId=n_id
                )
            notice_end()


@task(klass=AwsTask)
def tgw_attachments(ctx, session=None, **kwargs):
    """
    lists all tgw attachments
    """
    ec2 = session.client('ec2')
    attachments = ec2.describe_transit_gateway_attachments()
    attachments = attachments['TransitGatewayAttachments']
    attachments = [ x for x in attachments if x['State'] != 'deleted' ]
    header = [ 'id', 'name', 'route_table' ]
    rows = []
    for x in attachments:
        rows.append([
            x['TransitGatewayAttachmentId'],
            name_tag(x),
            x.get('Association', {}).get('TransitGatewayRouteTableId') or ''
        ])
    rows.sort(key=lambda lx: (lx[1], lx[0]))
    print()
    print_table(header, rows)


@task(klass=AwsTask)
def associate_tgw_attachment(ctx, attachment_id, route_table_id, session=None, **kwargs):
    """
    moves the transit gateway attachment to the specified route table
    """
    ec2 = session.client('ec2')
    attachments = ec2.describe_transit_gateway_attachments(
        TransitGatewayAttachmentIds=[ attachment_id ],
    )
    attachments = attachments['TransitGatewayAttachments']
    if not attachments:
        notice_end('not found')
        return
    x = attachments[0]
    current_route_table_id = x.get('Association', {}).get('TransitGatewayRouteTableId') or ''
    if current_route_table_id == route_table_id:
        notice_end('already associated')
        return
    if current_route_table_id:
        notice(f'disassociating from {current_route_table_id}')
        ec2.disassociate_transit_gateway_route_table(
            TransitGatewayRouteTableId=current_route_table_id,
            TransitGatewayAttachmentId=attachment_id,
        )
        notice_end()
    notice(f'associating to {route_table_id}')
    time.sleep(30)
    ec2.associate_transit_gateway_route_table(
        TransitGatewayRouteTableId=route_table_id,
        TransitGatewayAttachmentId=attachment_id,
    )
    notice_end()



from datetime import datetime
import time
from typing import Any

import pytz
from halo import Halo
from raft.tasks import task
from termcolor import cprint
from ...base.utils import notice, notice_end, get_context_value
from ...base.utils import print_table
from ..base import AwsTask


__all__ = [
    'build',
    'build_status',
    'build_logs',
    'stop_build',
    'running_builds',
    'show_phases',
    'recent_builds',
]


def as_bool(st: Any) -> bool:
    """
    converts a string or int to a boolean true / false value.  accepts
    `yes`, `true`, and `1` as true
    """
    if isinstance(st, str):
        return st.lower() in ( 'yes', 'true', '1' )
    return bool(st)


def yield_projects(client=None, session=None):
    if not client:
        client = session.client('codebuild')
    paginator = client.get_paginator('list_projects')
    for x in paginator.paginate():
        yield from x['projects']


def project_exists(project, client=None, session=None):
    notice(f'checking {project}')
    projects = list(yield_projects(client, session))
    if project in projects:
        notice_end()
        return True
    notice_end('not found')
    return False


@task(klass=AwsTask, help={
    'project': 'the name of the codebuild project',
    'branch': 'the git branch name that we want to build',
    'follow': 'specify this flag to follow the build after we kick it off',
    'verbose': 'specify this flag for more verbose output',
})
def build(ctx, project=None, branch=None, follow=True, verbose=False,
          session=None, **kwargs):
    """
    Kicks off a build of the specified project for the given branch

    e.g.,  convoke aws.build project1 -b branch1

    :param ctx:      the raft context
    :param project:  the name of the codebuild project
    :param branch:   the branch name in git or bitbucket
    :param follow:   tail the build logs after kicking off
    :param verbose:  verbose build log output
    :param session:  don't fill this out, it will be filled in for us by
        `AwsTask`

    configuration:

    ```yaml
    codebuild:
      project: my_codebuild_project
      default_branch: some_branch
    ```
    """
    client = session.client('codebuild')
    project = project or get_context_value(ctx, 'codebuild.project')
    key = 'codebuild.default_branch'
    branch = branch or get_context_value(ctx, key, 'master')
    if not project:
        notice_end('please specify a project')
        return
    if project_exists(project, client):
        notice('kicking off build')
        client.start_build(
            projectName=project,
            sourceVersion=branch,
        )
        notice_end()
    if as_bool(follow):
        notice('waiting 10 seconds for build to start')
        time.sleep(10)
        notice_end()
        build_logs(ctx, project, verbose=verbose, session=session)


def latest_build(project, client=None, session=None):
    notice('getting latest build id')
    client = client or session.client('codebuild')
    result = client.list_builds_for_project(
        projectName=project, sortOrder='DESCENDING')
    for x in result['ids']:
        notice_end(f'{x}')
        return x
    notice_end('waiting for godot')
    return None


def get_running_builds(project, client=None, session=None, **kwargs):
    notice('checking running builds')
    client = client or session.client('codebuild')
    result = client.list_builds_for_project(
        projectName=project, sortOrder='DESCENDING')
    build_ids = result['ids']
    result = client.batch_get_builds(ids=build_ids)
    running_build_ids = []
    for x in result['builds']:
        if x['buildStatus'] == 'IN_PROGRESS':
            running_build_ids.append(x['id'])
    if running_build_ids:
        n = len(running_build_ids)
        notice_end(f'{n} running build{"s" if n != -1 else ""}')
        return running_build_ids
    notice_end('waiting for godot')
    return []


@task(klass=AwsTask, help={
    'project': 'the codebuild project name',
    'n_max': 'the maximum number of recent builds to list',
})
def recent_builds(ctx, project=None, n_max=10, client=None,
                  session=None, **kwargs):
    """
    shows the n_max most recent builds in codebuild for this project
    """
    project = project or get_context_value(ctx, 'codebuild.project')
    if not project:
        notice_end('please specify a project')
        return
    notice('checking recent builds')
    client = client or session.client('codebuild')
    result = client.list_builds_for_project(
        projectName=project,
        sortOrder='DESCENDING')
    notice_end()
    build_ids = result['ids'][:n_max]
    notice('getting build details')
    result = client.batch_get_builds(ids=build_ids)
    notice_end()
    notice('keys')
    rows = []
    for i, x in enumerate(result['builds'], 1):
        if i == 1:
            notice_end(x.keys())
        rows.append((
            x['id'],
            x['buildStatus'],
            x['startTime'].isoformat(' ', 'seconds')))
    print_table([ 'build_id', 'status', 'started'], rows)


@task(klass=AwsTask, help={
    'project': 'name of the codebuild project',
    'session': 'do not use',
})
def running_builds(ctx, project=None, session=None, **kwargs):
    """
    prints a list of running build ids for a codebuild project
    """
    project = project or get_context_value(ctx, 'codebuild.project')
    if not project:
        notice_end('please specify a project to build')
        return
    build_ids = get_running_builds(project, session=session)
    if build_ids:
        print_table([ 'build_id' ], [ build_ids ])
    else:
        cprint('no running builds\n', 'blue')


@task(klass=AwsTask, help={
    'project': 'name of the codebuild project, optional',
    'build_id': 'the id of the build if you have it handy',
    'session': 'do not use',
})
def show_phases(ctx, project=None, build_id=None, client=None, session=None,
                **kwargs):
    """
    Shows the phases and their timings for a particular codebuild build
    :param str project: the name of the codebuild project
    :param raft.context.Context ctx: the convocation context
    :param str build_id: the build id of the codebuild build
    :param client: the boto3 codebuild client, if you have it handy
    :param boto3.Session session: the boto3 session, if you have it handy
    """
    client = client or session.client('codebuild')
    if not build_id:
        project = project or get_context_value(ctx, 'codebuild.project')
        if not project:
            notice_end('please specify a project')
            return
    build_id = build_id or latest_build(project, client, session)
    if build_id is None:
        notice_end('nothing to be done')
        return
    notice('getting build details')
    result = client.batch_get_builds(ids=[ build_id ])
    data = result['builds'][0]
    notice_end()
    phases = data['phases']
    fields = [ 'phaseType', 'phaseStatus', 'durationInSeconds' ]
    header = [ 'type', 'status', 'time' ]
    total_time = 0
    rows = []
    for x in phases:
        rows.append([ f'{x.get(field)}'.lower() for field in fields ])
        total_time += x.get('durationInSeconds', 0)
    rows.append([ '-----', '------', '-----' ])
    rows.append([ 'total time', '', f'{total_time}' ])
    print_table(header, rows)


def is_done(build_id, client, session=None) -> bool:
    """
    returns true if the codebuild build is done
    :param build_id: the build id in codebuild
    :param client: the boto3 codebuild client if we have it handy
    :param session: the boto3 session
    """
    client = client or session.client('codebuild')
    result = client.batch_get_builds(ids=[ build_id ])
    data = result['builds'][0]
    return data['buildStatus'] != 'IN_PROGRESS'


def format_log_line(message, verbose=False):
    if message.startswith('[Container]'):
        message = message[12:]
        dt = message[:19].strip()
        dt = datetime.strptime(dt, '%Y/%m/%d %H:%M:%S')
        dt = dt.replace(tzinfo=pytz.UTC)
        dt = dt.astimezone(pytz.timezone('CST6CDT'))
        message = message[20:]
        if message.startswith('Running command '):
            cprint(f"[{dt.strftime('%Y-%m-%d %H:%M:%S')}] ", 'cyan')
            message = f'----- {message[16:].strip()} -----'
            cprint(message, 'magenta')
            print()
        elif verbose:
            cprint(f"[{dt.strftime('%Y-%m-%d %H:%M:%S')}] ", 'cyan')
            print(message.strip())
        return
    print(message, end='')


def get_build_logs(project, build_id, start_time=None,
                   spinner=None, verbose=False, session=None):
    """
    codebuild stores logs in `/aws/codebuild/project`.  this function
    gets the build logs for the specified project and build_id from the
    specified start time, and formats the lines for output
    """
    from ..cloudwatch import yield_logs
    group_name = f'/aws/codebuild/{project}'
    stream_name = build_id.split(':')[-1]
    lines = list(yield_logs(group_name, stream_name, start_time, session))
    lines.sort(key=lambda lx: lx['timestamp'])
    for x in lines:
        if spinner:
            spinner.stop()
        format_log_line(x['message'], verbose)
    last_time = None
    if lines:
        last_time = lines[-1]['timestamp']
    return last_time


@task(klass=AwsTask, help=dict(
    build_id='optional, the build id of the codebuild project',
    project='optional, the name of the project',
), positional=[ 'project' ])
def build_status(ctx, project=None, build_id=None, session=None, **kwargs):
    """
    Shows the build status of the latest build of a codebuild project
    """
    client = session.client('codebuild')
    project = project or get_context_value(ctx, 'codebuild.project')
    if not project:
        notice_end('please specify a project')
        return
    build_id = build_id or latest_build(project, client)
    if build_id:
        show_phases(build_id, client)


@task(klass=AwsTask, help={
    'project': 'name of the codebuild project',
    'session': 'do not use',
})
def build_logs(ctx, project=None, build_id=None, verbose=False, session=None, **kwargs):
    """
    Tails the build logs of a codebuild project
    :param raft.context.Context ctx: the convocation context
    :param project: the name of the codebuild project
    :param build_id: the id of the codebuild build
    :param verbose: specify true / false for verbose logs
    :param session: don't fill this parameter in, it is injected
    """
    client = session.client('codebuild')
    project = project or get_context_value(ctx, 'codebuild.project')
    if not project:
        notice_end('please specify a project')
        return
    logs = session.client('logs')
    if isinstance(verbose, str) and verbose.lower() in [ 'yes', 'true' ]:
        verbose = True
    if not build_id:
        build_ids = get_running_builds(project, client)
        if build_ids:
            build_id = build_ids[0]
        else:
            build_id = latest_build(project, client)
            if not build_id:
                return
    start_time = None
    spinner = Halo('waiting')
    following = False
    for _ in range(6):
        try:
            if not spinner.spinner_id:
                spinner.start()
            while not is_done(build_id, client):
                next_time = get_build_logs(
                    project, build_id, start_time,
                    spinner, verbose, session)
                if not spinner.spinner_id:
                    spinner.start()
                start_time = next_time or start_time
                time.sleep(1)
                following = True
            if following:
                spinner.succeed('finished building')
                with Halo('waiting 5 seconds for cloudwatch cool down') as s:
                    time.sleep(5)
                    s.succeed()
            get_build_logs(
                project, build_id, start_time,
                None, verbose, session)
            break
        except logs.exceptions.ResourceNotFoundException:
            spinner.stop()
            notice_end('no build logs found')
            with Halo('waiting 3 seconds before retry'):
                time.sleep(3)


@task(klass=AwsTask, help={
    'project_or_build_id':
        'name of the codebuild project or the id of the codebuild'
        ' build that you want to stop',
    'session': 'do not use',
})
def stop_build(ctx, project_or_build_id=None, session=None, **kwargs):
    """
    stops the most running, still running build, if one exists

    :param raft.context.Context ctx: the convocation context
    :param str project_or_build_id: the name of the codebuild project
    :param session: don't fill this parameter in, it is injected
    """
    client = session.client('codebuild')
    try:
        project_or_build_id = project_or_build_id or ctx.codebuild.project
    except AttributeError:
        notice_end('please specify a project or build_id')
        return
    project, build_id = project_or_build_id, None
    if ':' in project:
        build_id = project
        project = project.split(':')[0]
    build_ids = get_running_builds(project, client)
    for x in build_ids:
        if build_id and x != build_id:
            continue
        notice(f'stopping {x}')
        client.stop_build(id=x)
        notice_end()
        break



import copy

from raft import task
from ...base.utils import print_table, notice, notice_end
from ..base import AwsTask


def rule_to_yaml(rule, prefix='Source'):
    data = dict(IpProtocol=rule['IpProtocol'])
    if rule['IpProtocol'] != '-1':
        data['FromPort'] = rule.get('FromPort', -1)
        data['ToPort'] = rule.get('ToPort', -1)
        if rule['IpProtocol'].startswith('icmp') and data['FromPort'] == 0:
            data['FromPort'] = -1
    ipv4_ranges = rule.get('IpRanges') or []
    for x in ipv4_ranges:
        d = copy.copy(data)
        cidr = x.get('CidrIp')
        if cidr:
            d['CidrIp'] = cidr
        description = x.get('Description')
        if description:
            d['Description'] = description
        yield d
    ipv6_ranges = rule.get('Ipv6Ranges') or []
    for x in ipv6_ranges:
        d = copy.copy(data)
        cidr = x.get('CidrIpv6')
        if cidr:
            d['CidrIpv6'] = cidr
        description = x.get('Description')
        if description:
            d['Description'] = description
            yield d
    pairs = rule.get('UserIdGroupPairs') or []
    for pair in pairs:
        d = copy.copy(data)
        description = pair.get('Description')
        if description:
            d['Description'] = description
        d[f'{prefix}SecurityGroupId'] = pair['GroupId']
        owner_id = pair.get('UserId')
        if owner_id and prefix == 'Source':
            d[f'{prefix}SecurityGroupOwnerId'] = owner_id
        yield d


@task(klass=AwsTask)
def security_group_to_yaml(ctx, logical_name=None, group_id=None,
                           session=None, quiet=False, **kwargs):
    """
    generates the cloudformation yaml for a security group
    """
    ec2 = session.resource('ec2')
    group = ec2.SecurityGroup(group_id)
    ingress, egress = [], []
    for x in group.ip_permissions:
        data = rule_to_yaml(x)
        ingress.extend(data)
    for x in group.ip_permissions_egress:
        data = rule_to_yaml(x, 'Destination')
        egress.extend(data)
    properties = dict(
        GroupDescription=group.description,
        GroupName=group.group_name,
        VpcId=group.vpc_id,
        Tags=group.tags,
        SecurityGroupIngress=ingress,
        SecurityGroupEgress=egress,
    )
    properties = { k: v for k, v in properties.items() if v is not None }
    data = {
        logical_name: dict(
            Type='AWS::EC2::SecurityGroup',
            DeletionPolicy='Retain',
            Properties=properties,
        ),
    }
    if not quiet:
        from convocations.base.utils import dump_yaml
        dump_yaml(data)
    return data


@task(klass=AwsTask)
def import_security_group(ctx, logical_name, group_id, filename,
                          skip_review=False, session=None, prefix=None,
                          **kwargs):
    """
    creates a yaml template for a security group and imports it into cloudformation

    if skip_review is set, we will immediately try to execute the change set
      in cloudformation
    """
    from .import_resource_to_cfn import import_resource_to_cfn
    data = security_group_to_yaml(
        ctx, logical_name, group_id, session=session, quiet=True)
    resources_to_import = [dict(
        ResourceType='AWS::EC2::SecurityGroup',
        LogicalResourceId=list(data.keys())[0],
        ResourceIdentifier=dict(GroupId=group_id),
    )]
    import_resource_to_cfn(
        data, resources_to_import,
        filename, None, skip_review, session, prefix)


@task(klass=AwsTask)
def security_groups(ctx, session=None, **kwargs):
    from ..ec2 import get_tag
    notice('loading security groups')
    ec2 = session.client('ec2')
    paginator = ec2.get_paginator('describe_security_groups')
    groups = []
    for page in paginator.paginate():
        groups += page['SecurityGroups']
    notice_end()
    header = [ 'id', 'name', 'tag_name', 'description' ]
    rows = []
    for group in groups:
        row = [
            group['GroupId'], group.get('GroupName', '')[:30],
            get_tag(group, 'name') or '', group.get('Description', '')[:30],
        ]
        rows.append(row)
    rows.sort(key=lambda lx: (lx[1] or lx[2]))
    print_table(header, rows)



# pylint: disable=not-context-manager
import contextlib
import os
from tempfile import NamedTemporaryFile
from raft.tasks import task

from ...base.utils import notice, notice_end
from ..base import AwsTask
from . import instance_by_name
from . import get_tag


@contextlib.contextmanager
def download_deploy_key(session, parameter_name):
    """
    downloads a deploy key from ssm to a temporary file, yields the temporary
    filename context
    :param session: the aws session
    :type  session: boto3.session.Session
    :param parameter_name: the name of the ssm parameter
    :type  parameter_name: str
    :return:
    """
    ssm = session.client('ssm')
    if not parameter_name.startswith('/'):
        parameter_name = f'/{parameter_name}'
    parameter = ssm.get_parameter(Name=parameter_name, WithDecryption=True)
    with NamedTemporaryFile(mode='w', buffering=1) as f:
        f.write(parameter['Parameter']['Value'])
        f.flush()
        os.chmod(f.name, 0o600)
        yield f


@task(klass=AwsTask, help={
    'public': 'use this flag to ssh to the public ip address instead'
              ' of the private ip address',
    'username': 'specify an alternative username to use when logging'
                ' into the remote server--e.g., centos, ec2-user or ubuntu.',
    'name': 'the tag name to match in aws.  n.b., we will use the first'
            ' instance with that name that appears in the api response.  you'
            ' may want to use the `instances` convocation to narrow down'
            ' conflicts'
})
def ssh(ctx, name, username=None, public=False, port=22, ip_override=None,
        session=None, **kwargs):
    """
    allows us to ssh to the instance if the key_pair was created by the
    new_key_pair convocation

    First, a new keypair is generated using the
    `new_key_pair` convocation and saved as namespace/name_deploy_key.
    This will create a new ec2 ssh key pair and store
    the private key in ssm under the path /namespace/name_deploy_key.  The
    deploy key is encrypted using the cmk with the namespace alias.  Then,
    when the instance is created, the key pair will be specified for the
    ssh instance.
    """
    x = instance_by_name(ctx, name, session=session)
    notice('getting key name')
    key_name = get_tag(x, 'deploy_key') or x.key_name
    notice_end(key_name)
    notice('username')
    username = username or get_tag(x, 'ssh_user')
    notice_end(username)
    notice('getting ip address')
    if public:
        ip_address = ip_override or x.public_ip_address or x.private_ip_address
    else:
        ip_address = ip_override or x.private_ip_address
    notice_end(ip_address)
    parameter_name = key_name
    if not key_name.startswith('/'):
        parameter_name = f'/{key_name}'
    with download_deploy_key(session, parameter_name) as f:
        if username:
            ssh_cmd = f'ssh -i {f.name} -p {port} {username}@{ip_address}'
            notice_end(ssh_cmd)
            ctx.run(ssh_cmd, pty=True)
        else:
            ctx.run(f'ssh -i {f.name} -p {port} {ip_address}', pty=True)



# pylint: disable=redefined-builtin
from raft.tasks import task

from ...base.utils import print_table
from ..base import AwsTask
from . import get_tag


@task(klass=AwsTask, help=dict(
    id='[optional] a comma delimited list of ids',
))
def route_tables(ctx, id=None, session=None, **kwargs):
    """
    prints a list of all routes in the route tables specified in the `id` argument.
    route tables can be specified as a comma-delimited list.  if no ids
    are specified, we will display all routes in all route tables.
    """
    ec2 = session.client('ec2')
    if id:
        id = id.split(',')
        rg = ec2.describe_route_tables(RouteTableIds=id)
    else:
        rg = ec2.describe_route_tables()
    header = [ 'id', 'name', 'dest', 'next_hop', ]
    rows = []
    for x in rg['RouteTables']:
        for route in x['Routes']:
            rows.append([
                x['RouteTableId'],
                get_tag(x, 'Name') or '',
                route.get('DestinationCidrBlock') or route.get('DestinationIpv6CidrBlock') or '',
                route.get('GatewayId') or route.get('InstanceId') or
                route.get('NetworkInterfaceId') or route.get('TransitGatewayId') or
                route.get('VpcPeeringConnectionId') or ''
            ])
    print_table(header, rows)



import os
import sys
from raft import task


@task
def global_python(ctx, pythons=None):
    """
    finds a global python interpreter in order of preference
    :param ctx:
    :param pythons:
    :return:
    """
    possible = pythons or [
        'python3.10',
        'python3.9',
        'python3.8',
        'python3.7',
        'python3.6',
    ]
    for x in possible:
        result = ctx.run(f'which {x}', warn=True)
        if result.ok:
            return result.stdout.strip()
    for x in possible:
        st = x.replace('.', '')
        st = os.path.join('/', st, 'python.exe')
        if os.path.exists(st):
            return st
    return None


@task
def setup(ctx, name=None):
    """
    creates the virtual environment and installs prerequisites
    """
    from ..aws.codebuild.detect import is_codebuild
    if is_codebuild():
        from ..aws.codebuild import setup as cb_setup
        cb_setup(ctx)
        return

    default_name = '.'
    if sys.platform in ('linux', 'darwin'):
        default_name = ctx.venv_name if 'venv_name' in ctx else '.'
    elif sys.platform == 'win32':
        default_name = ctx.windows_venv_name if 'windows_venv_name' in ctx else '.wenv'
    name = name or default_name
    expected_python = os.path.join(name, 'bin', 'python')
    if not os.path.exists(expected_python):
        python = global_python(ctx)
        ctx.run(f'{python} -m venv {name}')
    if os.path.exists(expected_python):
        python = expected_python
        pip = f'{python} -m pip install'
        result = ctx.run(f'{pip} -U wheel', pty=True)
        if result.exited:
            return
        result = ctx.run(f'{pip} -U pip setuptools raft', pty=True)
        if result.exited:
            return
        result = ctx.run(f'{pip} -r requirements.txt', pty=True)
        if result.exited:
            return
        if os.path.exists('setup.py'):
            result = ctx.run(f'{pip} -e .', pty=True)
            if result.exited:
                return
    else:
        print('no suitable python version found')



import time
from raft import task
from ..base import AwsTask


__all__ = [
    'launch_test_instance',
]


@task(klass=AwsTask)
def launch_test_instance(ctx, source_server, profile=None, session=None, **kwargs):
    """
    launches a source server in test mode
    """
    from boto3.session import Session
    from ...base.utils import notice, notice_end, print_table
    session = session or Session(profile_name=profile)
    mgn = session.client('mgn')
    paginator = mgn.get_paginator('describe_source_servers')
    source_servers = {}
    source_server = source_server.lower()
    notice('loading source servers')
    for page in paginator.paginate():
        for x in page['items']:
            name = x['sourceProperties']['identificationHints']['hostname']
            st_id = x['sourceServerID']
            source_servers[name.lower()] = st_id
    if source_server not in source_servers:
        notice_end(f'{source_server} was not found')
        header = [ 'name', 'id' ]
        rows = [ [ key, value ] for key, value in source_servers.items() ]
        print_table(header, rows)
        return
    st_id = source_servers[source_server]
    notice_end(st_id)
    notice(f'starting test of {st_id}')
    response = mgn.start_test(sourceServerIDs=[ st_id ])
    job_id = response['job']['jobID']
    notice_end(job_id)
    token = None
    conversion_server_id = None
    for y in range(11):
        notice(f'wait {y + 1}')
        response = mgn.describe_job_log_items(jobID=job_id, nextToken=token)
        token = response['nextToken']
        for x in response['items']:
            conversion_server_id = x['eventData'].get('conversionServerID')
            if conversion_server_id:
                break
        if conversion_server_id:
            break
        notice_end()
        time.sleep(15)
    if conversion_server_id:
        notice('modifying termination')
        ec2 = session.client('ec2')
        ec2.modify_instance_attribute(
            Attribute='disableApiTermination',
            DisableApiTermination=dict(Value=True),
            InstanceId=conversion_server_id,
        )
        notice_end()



# pylint: disable=W0603
from contextlib import nullcontext
import logging
from logging.config import dictConfig
import os
import platform
import sys
from threading import Lock
from raft import task


g_settings = None
settings_lock = Lock()
__all__ = [
    'which_python',
    'which_branch',
    'setup_django',
    'g_settings',
    'settings_lock',
    'setup_logging',
    'debug_logging',
    'activation_script',
    'is_windows',
    'is_ubuntu',
    'is_wsl',
    'is_aws',
    'activation_context',
    'which_docker_compose',
    'django_it',
]


def which_python(ctx):
    """
    Used to determine which python version is available
    """
    python = None
    pythons = [ 'python3.10', 'python3.9', 'python3.8', 'python3.7', 'python3.6' ]
    for x in pythons:
        result = ctx.run(f'which {x}')
        if result.exited == 0:
            python = result.stdout.strip()
            break
    if not python:
        print('could not find a suitable python interpreter')
        sys.exit(0)
    return python


def which_branch():
    """
    used to determine the branch being run in codebuild, if any
    """
    branch = os.environ.get('CODEBUILD_WEBHOOK_TRIGGER') or ''
    return branch.split('/')[-1]


@task
def setup_django(ctx):
    """ bring up django app settings """
    django_it(ctx)


def django_it(ctx, logging_type=None):
    from ..base.utils import get_context_value
    try:
        import django
        settings_module = get_context_value(ctx, 'django.settings')
        if logging_type:
            os.environ.setdefault('TUNDRA_LOGGING', logging_type)
        os.environ.setdefault('DJANGO_SETTINGS_MODULE', settings_module)
        django.setup()
    except ImportError:
        pass


def load_yaml(f):
    """
    f can be a filename, a string of yaml, or a buffer
    """
    from ..base.utils import notice_end
    try:
        import yaml
        try:
            with open(f, 'r', encoding='utf-8') as g:
                return yaml.load(g, yaml.SafeLoader)
        except (FileNotFoundError, OSError, TypeError):
            return yaml.load(f, yaml.SafeLoader)
    except ImportError:
        notice_end('could not find yaml')
        sys.exit(1)


@task
def setup_logging(ctx, pet=None, which='debug'):
    """
    setup logging by reading in the config

    ```yaml
    tundra:
      logging_configs:
        debug: conf/logging.debug.yml
        prod: conf/logging.prod.yml
      pet: sesame_meow_cat
    ```
    """
    from ..base.utils import get_context_value, notice_end
    configs = get_context_value(ctx, 'tundra.logging_configs') or {}
    filename = configs.get(which)
    if not os.path.exists(filename):
        notice_end(f'{filename} was not found')
        return
    dictConfig(load_yaml(filename))
    log = logging.getLogger(__name__)
    pet = pet or ctx.tundra.pet
    if pet:
        log.info('Hello, %s!', pet)


@task
def debug_logging(ctx, pet=None):
    """
    setup logging by reading in the config
    """
    setup_logging(ctx, pet, 'debug')


def activation_script():
    x = platform.uname()
    if x.system.lower() == 'windows':
        venv_dir = '.wenv'
        return '. ' + os.path.join(venv_dir, 'scripts', 'activate.ps1')
    return 'source bin/activate'


def is_ubuntu():
    """
    returns true if we are in an ubuntu environment
    """
    x = platform.uname()
    x = 'ubuntu' in x.version.lower()
    return x or 'ubuntu' in platform.platform().lower()


def is_aws() -> bool:
    """
    returns true if the platform uname indicates an aws image

    if an ami is running in aws, the release will look something like:
    `5.4.0-1089-aws`
    """
    x = platform.uname()
    return 'aws' in x.release.lower()


def is_wsl() -> bool:
    """
    returns true if we are in wsl v2.

    wsl2 images will have a uname like

    ```
    Linux e 5.10.102.1-microsoft-standard-WSL2 #1 SMP Wed Mar 2 00:30:59 UTC 2022 x86_64 x86_64 x86_64 GNU/Linux
    ```
    """
    x = platform.uname()
    return x.system.lower() == 'linux' and 'wsl2' in x.release.lower()


def is_wsl1() -> bool:
    """
    returns true if we are in wsl v1

    wsl1 images will have a uname like

    ```
    Linux e 4.4.0-19041-Microsoft #1237-Microsoft Sat Sep 11 14:32:00 PST 2021 x86_64 x86_64 x86_64 GNU/Linux
    ```

    :return (bool):
    """
    x = platform.uname()
    boo = x.system.lower() == 'linux'
    boo = boo and x.release.lower().endswith('microsoft')
    boo = boo and not is_wsl()
    return boo


def is_windows():
    x = platform.uname()
    return x.system.lower() == 'windows'


def activation_context(ctx, codebuild=False):
    from ..aws.codebuild import detect
    codebuild = codebuild or detect.is_codebuild()
    if not codebuild:
        return ctx.prefix(activation_script())
    return nullcontext()


def which_docker_compose():
    if sys.platform == 'win32':
        return 'docker-compose.exe'
    return 'docker-compose'



import os
import posixpath

from raft import task


names = [ 'cscshared/{{ name }}', ]


def short_name():
    """
    returns the shortname for the docker image

    for example, the shortname for namespace/short_name:tag will be `short_name`
    """
    for name in names:
        name = posixpath.basename(name)
        name = name.split(':', 1)[0]
        return name


def local_tag():
    return f'{short_name()}:local'


@task
def build(ctx):
    """
    builds the docker image
    """
    ctx.run(f'docker build -t {local_tag()} .')


@task
def bump_version(ctx, which=None):
    """
    bumps the version for publishing
    """
    if 'CODEBUILD_BUILD_ID' in os.environ:
        print('in codebuild, skipping')
        return
    if not which:
        minor = version().rsplit('.', 2)[-1]
        which = 'major' if minor == '9' else 'minor'
    ctx.run(f'bumpversion {which}')
    ctx.run('git push --all')


def version():
    with open('version.txt', 'r') as f:
        st = f.read()
        st = st.strip()
    return st


def tags():
    for name in names:
        for x in [ version(), 'latest' ]:
            yield f'{name}:{x}'


@task
def tag(ctx):
    """
    tags the build with the appropriate version(s)
    """
    for image in tags():
        print(f'tagging {image}')
        ctx.run(f'docker tag {local_tag()} {image}')


def image_exists(ctx, image):
    """
    uses the docker manifest cli to determine if an image is available in the
    remote
    """
    result = ctx.run(f'docker manifest inspect {image}', warn=True, hide=True)
    return not bool(result.return_code)


@task()
def publish(ctx, profile=None, local=False):
    """
    publishes the image to docker hub and ecr as needed
    """
    if 'CODEBUILD_BUILD_ID' not in os.environ and not local:
        bump_version(ctx)
        return
    build(ctx)
    bump_version(ctx)
    tag(ctx)
    if local:
        profile_arg = f'--profile={profile}' if profile else ''
        d = os.getcwd()
        ctx.run(f"eval $({d}/venv/bin/aws "
                f"{profile_arg} ecr get-login --no-include-email)")
    for image in tags():
        if not image_exists(ctx, image):
            print(f'pushing {image}')
            ctx.run(f'docker push {image}')



from raft.collection import Collection
from raft.program import Program
from .files import upload_files
from .permissions import grant_access
from .permissions import check_access
from ..version import version


ns = Collection(grant_access, check_access, upload_files)
program = Program(version=version, namespace=ns)



from raft.tasks import task
from convocations.base.utils import print_table
from .base import CheckpointTask


@task(klass=CheckpointTask)
def layers(ctx, client=None, host=None, **kwargs):
    """
    lists all policy layers on the checkpoint by name
    """
    from cpapi import APIResponse
    result: APIResponse = client.api_call('show-access-layers')
    rg = result.data['access-layers']
    names = [ [ x['name'] ] for x in rg ]
    header = [ 'name' ]
    print_table(header, names)
    return rg



from raft.tasks import task
from ..base import AwsTask
from ...base.utils import print_table, notice, notice_end


@task(klass=AwsTask)
def run_posh(ctx, name, filename, session=None, **kwargs):
    """
    runs a command on the specified windows instance using ssm run command
    :param ctx: the raft / convoke context
    :param str name: the name or instance id of the ec2 instance
    :param str filename: the /path/to the .ps1 script that you want to run
    :param session: leave this blank
    """
    from ..ec2 import yield_instances
    from ..base import get_tag
    notice(f'loading {filename}')
    with open(filename, 'rt') as f:
        commands = f.readlines()
    instances = list(yield_instances(name, session))
    instance_ids = [ x['InstanceId'] for x in instances ]
    header = [ 'instance_id', 'name' ]
    rows = []
    for x in instances:
        rows.append([ x['InstanceId'], get_tag(x, 'name') ])
    print()
    print_table(header, rows)
    print()
    notice(f'running command on {len(rows)} instances')
    ssm = session.client('ssm')
    result = ssm.send_command(
        InstanceIds=instance_ids,
        DocumentName='AWS-RunPowerShellScript',
        Parameters=dict(commands=commands)
    )
    result = result['Command']
    command_id = result['CommandId']
    notice_end(command_id)
    waiter = ssm.get_waiter('command_executed')
    for instance_id in instance_ids:
        notice(f'waiting for {instance_id}')
        waiter.wait(
            CommandId=command_id,
            InstanceId=instance_id,
            WaiterConfig=dict(Delay=15))
        notice_end()
        x = ssm.get_command_invocation(CommandId=command_id, InstanceId=instance_id)
        print(x['StandardOutputContent'])
        print(x['StandardErrorContent'])



from raft.tasks import task
from convocations.base.utils import notice, notice_end
from .base import CheckpointTask


@task(klass=CheckpointTask)
def api_version(ctx, client=None, host=None, **kwargs):
    """
    lists the current api version
    """
    from cpapi import APIResponse
    result: APIResponse = client.api_call('show-api-versions')
    notice('checkpoint api version')
    notice_end(result.data['current-version'])



from raft.tasks import task
from convocations.base.utils import print_table, notice, notice_end
from .base import CheckpointTask


@task(klass=CheckpointTask)
def services(ctx, exp='', protocol='tcp', host=None, client=None, **kwargs):
    """
    filters available services by expression
    """
    from cpapi.api_response import APIResponse
    payload = dict(limit=500)
    command = f'show-services-{protocol}'
    exp = exp.lower()
    response: APIResponse = client.api_call(command, payload=payload)
    print(f'{response.data}')
    total = response.data['total']
    rg = response.data['objects']
    if total > 500:
        payload['offset'] = 0
        while total > 0:
            total -= 500
            payload['offset'] += 500
            response = client.api_call(command, payload)
            rg += response.data['objects']
    rg = filter(lambda lx: exp in lx['name'].lower(), rg)
    header = [ 'name', 'type', 'port' ]
    rows = [ [ x.get('name'), x.get('type'), x.get('port') ] for x in rg ]
    print_table(header, rows)
    return rg


@task(klass=CheckpointTask)
def service(ctx, name, groups=True, protocol='tcp',
            host=None, client=None, prefix='', quiet=False, **kwargs):
    """
    shows the available service named `name`
    """
    from cpapi.api_response import APIResponse
    command = f'show-service-{protocol}'
    payload = dict(name=name)
    if not quiet:
        notice(f'{prefix}loading service')
    response: APIResponse = client.api_call(command, payload=payload)
    data = response.data
    if not quiet:
        notice_end(data['name'])
        notice(f'{prefix}port')
        notice_end(data['port'])
        notice(f'{prefix}protocol')
        notice_end(data.get('protocol'))
        if groups:
            groups = data.get('groups') or []
            for group in groups:
                notice(f'{prefix}group')
                notice_end(group['name'])
    return data


@task(klass=CheckpointTask)
def service_group(ctx, name, host=None, client=None, **kwargs):
    """
    shows the available service group named `name`
    """
    from cpapi.api_response import APIResponse
    command = 'show-service-group'
    payload = dict(name=name)
    notice('loading service group')
    response: APIResponse = client.api_call(command, payload=payload)
    group_data = response.data
    notice_end(group_data['name'])
    rows = []
    members = group_data['members']
    print()
    header = [ 'name', 'port', 'protocol', 'type' ]
    for x in members:
        service_data = service(ctx, name=x['name'], client=client, quiet=True)
        rows.append([
            x['name'],
            service_data['port'],
            service_data['protocol'],
            service_data['type'].replace('service-', '') ])
        x.update(service_data)
    print_table(header, rows)
    return group_data



from raft.tasks import task
from ...base.utils import notice, notice_end, print_table
from ..base import yielder, AwsTask, name_tag, get_path
from ..ec2 import get_tag, yield_instances, yield_instances_by_id
from ..wait_helpers import wait


__all__ = [
    'register',
    'deregister',
    'lb_health',
]


def yield_lbs(name='', exact=False, session=None, prefix=None):
    """
    Yields all ec2 classic load balancers that have the `Name` that
    matches the provided name.  If `exact` is `True`, the
    instance's name must match the `name` parameter exactly.
    All matching is case-insensitive
    """
    from botocore.exceptions import ClientError

    if name:
        name = name.lower()
    for x in yielder('elb', 'describe_load_balancers', session=session):
        x['name'] = x['LoadBalancerName']
        l_name = x['name'].lower()
        if not name:
            yield x
        elif name in l_name:
            if not exact or name == l_name:
                yield x
    try:
        for x in yielder('elbv2', 'describe_target_groups', session=session):
            x['name'] = x['TargetGroupName']
            l_name = x['name'].lower()
            if not name:
                yield x
            elif name in l_name:
                if not exact or name == l_name:
                    yield x
    except ClientError as ex:
        if ex.response['Error']['Code'] != 'TargetGroupNotFound':
            raise


def yield_target_groups(name='', exact=True, session=None, prefix=None):
    """
    Yields all target groups that have `Name` that
    matches the provided name.  If `exact` is `True`, the
    instance's name must match the `name` parameter exactly.
    All matching is case-insensitive
    """
    if name:
        name = name.lower()
    client = session.client('elbv2')
    for x in yielder('elbv2', 'describe_target_groups', session=session):
        x['name'] = target_group = x['TargetGroupName'].lower()
        if not name:
            yield x
        elif name in target_group:
            if not exact or name == target_group:
                yield x
        arn = x['TargetGroupArn']
        tags = client.describe_tags(ResourceArns=[arn])
        if tags['TagDescriptions']:
            tags = tags['TagDescriptions'][0]
            x['name'] = target_group = name_tag(tags)
            if name in target_group:
                yield x


def _deregister(lb, instances, session=None):
    is_target_group = 'TargetGroupArn' in lb
    client = 'elbv2' if is_target_group else 'elb'
    elb = session.client(client)
    lb_name = lb['name']
    names = [ get_tag(instance, 'name') for instance in instances ]
    notice(f"disabling {','.join(names)} on {lb_name}")
    if is_target_group:
        arn = lb['TargetGroupArn']
        instance_ids = [{ 'Id': x['InstanceId'] } for x in instances ]
        kwargs = dict(
            TargetGroupArn=arn,
            Instances=instance_ids,)
        elb.deregister_targets(**kwargs)
    else:
        instance_ids = [{ 'InstanceId': x['InstanceId'] } for x in instances ]
        kwargs = dict(
            LoadBalancerName=lb_name,
            Instances=instance_ids,)
        elb.deregister_instances_from_load_balancer(**kwargs)
    notice_end()
    waiter = 'target' if is_target_group else 'instance'
    waiter = elb.get_waiter(f'{waiter}_deregistered')
    wait(waiter, 'instances to deregister', **kwargs)


def _register(lb_or_target_group, instances, session=None):
    is_target_group = 'TargetGroupName' in lb_or_target_group
    client = session.client('elbv2' if is_target_group else 'elb')
    name = lb_or_target_group['name']
    arn = lb_or_target_group.get('TargetGroupArn')
    names = [ get_tag(instance, 'name') for instance in instances ]
    notice(f"enabling {','.join(names)} on {name}")
    if is_target_group:
        instance_ids = [{ 'Id': x['InstanceId'] } for x in instances ]
        kwargs = dict(
            TargetGroupArn=arn,
            Targets=instance_ids,
        )
        client.register_targets(**kwargs)
    else:
        instance_ids = [{ 'InstanceId': x['InstanceId'] } for x in instances ]
        kwargs = dict(
            LoadBalancerName=name,
            Instances=instance_ids,)
        client.register_instances_with_load_balancer(**kwargs)
    notice_end()
    if is_target_group:
        waiter = client.get_waiter('target_in_service')
    else:
        waiter = client.get_waiter('instance_in_service')
    wait(waiter, 'new instances to come into service', **kwargs)


def lookup_load_balancers(lb_name, session):
    notice(f'looking for load balancers {lb_name}')
    rg = list(yield_lbs(lb_name, session=session))
    if not rg:
        notice_end('not found')
        notice(f'looking for target groups {lb_name}')
        rg = list(yield_target_groups(lb_name, session=session))
        if not rg:
            notice_end('nothing found')
            return []
    names = [ x['name'] for x in rg ]
    notice_end(','.join(names))
    return rg


@task(klass=AwsTask)
def register(ctx, lb_name, instance_name, session=None, prefix=None):
    """
    Registers an instance with a load balancer or a target group

    Args:
         lb_name -- the name of the load balancer or target group
         instance_name -- the value of the name tag on the instance
    """
    rg = lookup_load_balancers(lb_name, session)
    notice(f'finding {instance_name}')
    instances = list(yield_instances(instance_name, session))
    names = [ get_tag(x, 'Name') for x in instances ]
    notice_end(','.join(names))
    _register(rg[0], instances, session)


@task(klass=AwsTask)
def deregister(ctx, lb_name, instance_name, session=None, prefix=None):
    """
    Deregisters an instance from a load balancer

    Args:

        ctx (raft.context.Context): the convocation context
        lb_name (str): name of the load balancer
        instance_name (str): name of the instance
        session (boto3.Session): don't specify this
        prefix: don't specify this either
    """
    load_balancers = lookup_load_balancers(lb_name, session)
    notice(f'finding {instance_name}')
    instances = list(yield_instances(instance_name, session))
    names = [ x['InstanceId'] for x in instances ]
    notice_end(','.join(names))
    _deregister(load_balancers[0], instances, session)


@task(klass=AwsTask)
def lb_health(ctx, name, session=None, **kwargs):
    """
    Displays the health of the targets on a load balancer
    """
    load_balancers = lookup_load_balancers(name, session)
    if not load_balancers:
        return
    header = [ 'lb', 'name', 'id', 'state',  ]
    rows = []
    for load_balancer in load_balancers:
        name = load_balancer['name']
        is_target_group = 'TargetGroupArn' in load_balancer
        client_name = 'elbv2' if is_target_group else 'elb'
        notice(f'checking health of {name}')
        client = session.client(client_name)
        if is_target_group:
            health_response = client.describe_target_health(
                TargetGroupArn=load_balancer['TargetGroupArn'],
            )
            health_response = health_response['TargetHealthDescriptions']
            id_key = 'Target.Id'
            state_key = 'TargetHealth.State'
            reason_key = 'TargetHealth.Reason'
        else:
            health_response = client.describe_instance_health(
                LoadBalancerName=load_balancer['LoadBalancerName'],
            )
            health_response = health_response['InstanceStates']
            id_key = 'InstanceId'
            state_key = 'State'
            reason_key = 'ReasonCode'
        instance_ids, ips = [], []
        for x in health_response:
            value = get_path(x, id_key)
            if value.startswith('i-'):
                instance_ids.append(value)
            else:
                ips.append(value)
        notice_end()
        notice('looking up instances')
        instances = {
            x['InstanceId']: x
            for x in yield_instances_by_id(session, instance_ids)
        }
        for x in ips:
            instances[x] = x
        notice_end()
        for x in health_response:
            instance_id = get_path(x, id_key)
            if instance_id.startswith('i-'):
                instance = instances[instance_id]
                instance_name = get_tag(instance, 'Name')
            else:
                instance_name = instance_id
            rows.append([
                name,
                instance_name,
                instance_id,
                get_path(x, state_key, ''),
            ])
    print_table(header, rows)



import os
from raft import task
from ..base.utils import notice_end
from ..aws.base import AwsTask


__all__ = [
    'staging',
    'prod',
    'dags',
]


@task(klass=AwsTask)
def staging(ctx, limit=None, branch=None, profile=None, session=None):
    """
    deploys to staging environment
    requires the following configuration:

    ```yaml
    aws:
      profile: profile_name
    tundra:
      deploy:
        deploy_keys:
          limit1: namespace/key1
          limit2: namespace/key2
        staging:
          limit: default_limit
          branch: staging
    ```
    """
    conf = ctx.tundra.deploy.staging
    limit = limit or conf.limit
    branch = branch or conf.branch
    _deploy(ctx, conf, limit, branch, session=session)


@task(klass=AwsTask)
def prod(ctx, limit=None, branch=None, profile=None, session=None, **kwargs):
    """
    deploys to the prod environment

    requires the following configuration:

    ```yaml
    aws:
      profile: profile_name
    tundra:
      deploy:
        prod:
          limit: default_limit
          branch: master
    ```
    """
    conf = ctx.tundra.deploy.prod
    limit = limit or conf.limit
    branch = branch or conf.branch
    _deploy(ctx, conf, limit, branch, session=session)


def _deploy(ctx, conf, limit, branch=None, playbook='deploy.yml', tags=None,
            session=None, **kwargs):
    with ctx.cd('deploy'):
        os.makedirs('.sockets', exist_ok=True)
        if os.path.exists('ansible.cfg'):
            os.environ['ANSIBLE_CONFIG'] = './ansible.cfg'
        ansible = f'ansible-playbook -i hosts {playbook}'
        if limit:
            ansible = f'{ansible} --limit {limit}'
        if branch:
            ansible = f'{ansible} -e branch={branch}'
        if tags:
            ansible = f'{ansible} --tags {tags}'
        if kwargs:
            extra = ' '.join([
                f'-e {key}="{value}"' for key, value in kwargs.items()
            ])
            ansible = f'{ansible} {extra}'
        ctx.run(ansible, pty=True)


@task(klass=AwsTask)
def flower(ctx, limit=None, profile=None, session=None):
    """
    deploys flower, requires the following configuration:

    ```yaml
    aws:
      profile: profile_name
    tundra:
      deploy:
        deploy_keys:
          limit1: namespace/key1
          limit2: namespace/key2
        flower:
          limit: default_limit
          playbook: flower.yml
    ```
    """
    conf = ctx.tundra.deploy.flower
    playbook = conf.get('playbook', 'flower.yml')
    limit = limit or conf.limit
    _deploy(ctx, conf, limit, None, playbook, session=session)


@task(klass=AwsTask)
def dags(ctx, dag, limit=None, profile=None, session=None):
    """
    pushes a specific dag from your local to the peaky / staging environment.
    this will overwrite any dags in these directories.

    requires the following context:

    ```yaml
    aws:
      profile: profile_name
    tundra:
      deploy:
        deploy_keys:
          limit1: namespace/key1
          limit2: namespace/key2
        dags:
          limit: default_limit
          playbook: dags.yml
    ```
    """
    conf = ctx.tundra.deploy.dags
    limit = limit or conf.get('limit', None)
    playbook = conf.get('playbook', 'dags.yml')
    if not limit:
        notice_end('no limit specified for dag deploys')
        return
    _deploy(ctx, conf, limit, None, playbook, session=session, dag=dag)
